package gregapi.util;

import static gregapi.data.CS.*;
import gregapi.GT_API;
import gregapi.block.IBlockExtendedMetaData;
import gregapi.block.IBlockPlacable;
import gregapi.block.IBlockTileEntity;
import gregapi.code.ArrayListNoNulls;
import gregapi.code.ItemStackContainer;
import gregapi.code.ItemStackMap;
import gregapi.code.TagData;
import gregapi.damage.DamageSources;
import gregapi.data.*;
import gregapi.data.CS.ItemsGT;
import gregapi.data.TC.TC_AspectStack;
import gregapi.enchants.Enchantment_Radioactivity;
import gregapi.item.IItemProjectile;
import gregapi.item.IItemUpdatable;
import gregapi.lang.LanguageHandler;
import gregapi.log.GT_Log;
import gregapi.network.packets.PacketSound;
import gregapi.old.BlockScanningEvent;
import gregapi.old.GT_Runnable_Sound;
import gregapi.old.GregTech_API;
import gregapi.old.PlayedSound;
import gregapi.old.interfaces.IDebugableBlock;
import gregapi.old.interfaces.tileentity.*;
import gregapi.old.items.GT_EnergyArmor_Item;
import gregapi.oredict.OreDictItemData;
import gregapi.oredict.OreDictManager;
import gregapi.oredict.OreDictMaterial;
import gregapi.oredict.OreDictMaterialStack;
import gregapi.recipes.GT_ModHandler;
import gregapi.tileentity.DelegatorTileEntity;
import gregapi.tileentity.ITileEntityDelegating;
import ic2.api.recipe.IRecipeInput;
import ic2.api.recipe.RecipeInputItemStack;
import ic2.api.recipe.RecipeInputOreDict;
import ic2.api.recipe.RecipeOutput;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.text.DateFormat;
import java.util.*;
import java.util.Map.Entry;

import net.minecraft.block.Block;
import net.minecraft.block.BlockFire;
import net.minecraft.block.material.Material;
import net.minecraft.enchantment.Enchantment;
import net.minecraft.enchantment.EnchantmentHelper;
import net.minecraft.entity.*;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.init.Blocks;
import net.minecraft.init.Items;
import net.minecraft.inventory.IInventory;
import net.minecraft.inventory.ISidedInventory;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.*;
import net.minecraft.nbt.NBTBase.NBTPrimitive;
import net.minecraft.network.play.server.S07PacketRespawn;
import net.minecraft.network.play.server.S1DPacketEntityEffect;
import net.minecraft.potion.Potion;
import net.minecraft.potion.PotionEffect;
import net.minecraft.server.MinecraftServer;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.tileentity.TileEntityChest;
import net.minecraft.util.AxisAlignedBB;
import net.minecraft.util.ChatComponentText;
import net.minecraft.util.ChunkCoordinates;
import net.minecraft.util.MathHelper;
import net.minecraft.world.World;
import net.minecraft.world.WorldServer;
import net.minecraft.world.biome.BiomeGenBase;
import net.minecraft.world.chunk.Chunk;
import net.minecraftforge.common.DimensionManager;
import net.minecraftforge.common.MinecraftForge;
import net.minecraftforge.common.util.ForgeDirection;
import net.minecraftforge.fluids.*;
import net.minecraftforge.fluids.FluidContainerRegistry.FluidContainerData;
import cpw.mods.fml.common.FMLCommonHandler;
import cpw.mods.fml.common.FMLLog;

/**
 * @author Gregorius Techneticies
 * 
 * Utility for accessing the random Utility Functions in a more short manner. The Short Name is for ease of overview and stands for "UtiliTy". :P
 */
public class UT {
	public static class Fluids {
		public static int id(FluidStack aFluid) {
			return aFluid.getFluid().getID();
		}
		
		public static boolean equal(FluidStack aFluid1, FluidStack aFluid2) {
			return equal(aFluid1, aFluid2, false);
		}
		
		public static boolean equal(FluidStack aFluid1, FluidStack aFluid2, boolean aIgnoreNBT) {
			return aFluid1 != null && aFluid2 != null && aFluid1.getFluid() == aFluid2.getFluid() && (aIgnoreNBT || ((aFluid1.tag == null) == (aFluid2.tag == null)) && (aFluid1.tag == null || aFluid1.tag.equals(aFluid2.tag)));
		}
		
	    public static ItemStack display(Fluid aFluid) {
	    	return aFluid==null?null:display(new FluidStack(aFluid, 0), F);
	    }
	    
	    public static ItemStack display(FluidStack aFluid, boolean aUseStackSize) {
	    	if (aFluid == null || aFluid.getFluid() == null) return null;
	    	ItemStack rStack = IL.Display_Fluid.getWithDamage(aUseStackSize ? aFluid.amount / 1000 : 1, id(aFluid));
	    	NBTTagCompound tNBT = new NBTTagCompound();
			tNBT.setLong("mFluidDisplayAmount", aFluid.amount);
			tNBT.setLong("mFluidDisplayHeat", aFluid.getFluid().getTemperature(aFluid));
			tNBT.setBoolean("mFluidState", aFluid.getFluid().isGaseous(aFluid));
			rStack.setTagCompound(tNBT);
	    	return rStack;
	    }
	    
	    private static Fluid WATER, DISTILLED_WATER, LAVA, MILK, STEAM;
	    
		/** @return if that Liquid is Water or Distilled Water */
		public static boolean water(FluidStack aFluid) {
			return aFluid != null && water(aFluid.getFluid());
		}
		
		/** @return if that Liquid is Water or Distilled Water */
		public static boolean water(Fluid aFluid) {
			if (aFluid			== null) return false;
			if (WATER			== null) WATER				= FluidRegistry.getFluid("water");
			if (DISTILLED_WATER	== null) DISTILLED_WATER	= FluidRegistry.getFluid("ic2distilledwater");
			return aFluid == WATER || aFluid == DISTILLED_WATER;
		}
		
		/** @return a Liquid Stack with given amount of Water. */
		public static FluidStack water(long aAmount) {
			if (WATER			== null) WATER				= FluidRegistry.getFluid("water");
			return new FluidStack(WATER, (int)aAmount);
		}
		
		/** @return if that Liquid is distilled Water */
		public static boolean distilledwater(FluidStack aFluid) {
			return aFluid != null && distilledwater(aFluid.getFluid());
		}
		
		/** @return if that Liquid is distilled Water */
		public static boolean distilledwater(Fluid aFluid) {
			if (aFluid			== null) return false;
			if (DISTILLED_WATER	== null) DISTILLED_WATER	= FluidRegistry.getFluid("ic2distilledwater");
			return aFluid == DISTILLED_WATER;
		}
		
		/** @return a Liquid Stack with given amount of distilled Water. */
		public static FluidStack distilledwater(long aAmount) {
			if (DISTILLED_WATER	== null) DISTILLED_WATER	= FluidRegistry.getFluid("ic2distilledwater");
			return new FluidStack(DISTILLED_WATER, (int)aAmount);
		}

		/** @return if that Liquid is distilled Water */
		public static boolean lava(FluidStack aFluid) {
			return aFluid != null && lava(aFluid.getFluid());
		}
		
		/** @return if that Liquid is Lava */
		public static boolean lava(Fluid aFluid) {
			if (aFluid			== null) return false;
			if (LAVA			== null) LAVA				= FluidRegistry.getFluid("lava");
			return aFluid == LAVA;
		}
		
		/** @return a Liquid Stack with given amount of distilled Water. */
		public static FluidStack lava(long aAmount) {
			if (LAVA		== null) LAVA					= FluidRegistry.getFluid("lava");
			return new FluidStack(LAVA, (int)aAmount);
		}
		
		/** @return if that Liquid is distilled Water */
		public static boolean steam(FluidStack aFluid) {
			return aFluid != null && steam(aFluid.getFluid());
		}
		
		/** @return if that Liquid is distilled Water */
		public static boolean steam(Fluid aFluid) {
			if (aFluid			== null) return false;
			if (STEAM			== null) STEAM				= FluidRegistry.getFluid("steam");
			return aFluid == STEAM;
		}
		
		/** @return a Liquid Stack with given amount of distilled Water. */
		public static FluidStack steam(long aAmount) {
			if (STEAM			== null) STEAM				= FluidRegistry.getFluid("steam");
			return new FluidStack(STEAM, (int)aAmount);
		}
		
		/** @return if that Liquid is distilled Water */
		public static boolean milk(FluidStack aFluid) {
			return aFluid != null && milk(aFluid.getFluid());
		}
		
		/** @return if that Liquid is distilled Water */
		public static boolean milk(Fluid aFluid) {
			if (aFluid			== null) return false;
			if (MILK			== null) MILK				= FluidRegistry.getFluid("milk");
			return aFluid == MILK;
		}
		
		/** @return a Liquid Stack with given amount of distilled Water. */
		public static FluidStack milk(long aAmount) {
			if (MILK			== null) MILK				= FluidRegistry.getFluid("milk");
			return new FluidStack(MILK, (int)aAmount);
		}
		
		public static String name(Fluid aFluid, boolean aLocalized) {
			if (aFluid == null) return "";
			String rName = aLocalized?aFluid.getLocalizedName(new FluidStack(aFluid, 0)):aFluid.getUnlocalizedName();
			if (rName.contains("fluid.") || rName.contains("tile.")) return Code.capitalise(rName.replaceAll("fluid.", "").replaceAll("tile.", ""));
			return rName;
		}
		
		public static String name(FluidStack aFluid, boolean aLocalized) {
			if (aFluid == null) return "";
			return name(aFluid.getFluid(), aLocalized);
		}
		
		public static FluidStack[] copyArray(FluidStack... aStacks) {
			FluidStack[] rStacks = new FluidStack[aStacks.length];
			for (int i = 0; i < aStacks.length; i++) if (aStacks[i] != null) rStacks[i] = aStacks[i].copy();
			return rStacks;
		}
		
		private static final List<FluidContainerData> sFluidContainerList = new ArrayListNoNulls<FluidContainerData>();
		private static final Map<ItemStackContainer, FluidContainerData> sFilledContainerToData = new ItemStackMap<ItemStackContainer, FluidContainerData>();
		private static final Map<ItemStackContainer, Map<Fluid, FluidContainerData>> sEmptyContainerToFluidToData = new ItemStackMap<ItemStackContainer, Map<Fluid, FluidContainerData>>();
		
		public static void reInit() {
			sFilledContainerToData.clear();
			sEmptyContainerToFluidToData.clear();
			for (FluidContainerData tData : sFluidContainerList) {
				sFilledContainerToData.put(new ItemStackContainer(tData.filledContainer), tData);
				Map<Fluid, FluidContainerData> tFluidToContainer = sEmptyContainerToFluidToData.get(new ItemStackContainer(tData.emptyContainer));
				if (tFluidToContainer == null) {
					sEmptyContainerToFluidToData.put(new ItemStackContainer(tData.emptyContainer), tFluidToContainer = new HashMap<Fluid, FluidContainerData>());
					GT_API.FLUIDMAPS.add(tFluidToContainer);
				}
				tFluidToContainer.put(tData.fluid.getFluid(), tData);
			}
		}
		
		public static void addFluidContainerData(FluidContainerData aData) {
			sFluidContainerList.add(aData);
			sFilledContainerToData.put(new ItemStackContainer(aData.filledContainer), aData);
			Map<Fluid, FluidContainerData> tFluidToContainer = sEmptyContainerToFluidToData.get(new ItemStackContainer(aData.emptyContainer));
			if (tFluidToContainer == null) {
				sEmptyContainerToFluidToData.put(new ItemStackContainer(aData.emptyContainer), tFluidToContainer = new HashMap<Fluid, FluidContainerData>());
				GT_API.FLUIDMAPS.add(tFluidToContainer);
			}
			tFluidToContainer.put(aData.fluid.getFluid(), aData);
		}
		
	    public static ItemStack fillFluidContainer(FluidStack aFluid, ItemStack aStack, boolean aRemoveFluidDirectly, boolean aCheckIFluidContainerItems) {
			if (Stacks.invalid(aStack) || aFluid == null) return null;
			if (water(aFluid) && Stacks.equal(aStack, new ItemStack(Items.glass_bottle, 1, CS.W))) {
				if (aFluid.amount >= 250) {
					if (aRemoveFluidDirectly) aFluid.amount -= 250;
					return new ItemStack(Items.potionitem, 1, 0);
				}
				return null;
			}
	    	if (aCheckIFluidContainerItems && aStack.getItem() instanceof IFluidContainerItem && ((IFluidContainerItem)aStack.getItem()).getFluid(aStack) == null && ((IFluidContainerItem)aStack.getItem()).getCapacity(aStack) <= aFluid.amount) {
				if (aRemoveFluidDirectly)
		    		aFluid.amount -= ((IFluidContainerItem)aStack.getItem()).fill(aStack = Stacks.copyAmount(1, aStack), aFluid, true);
				else
		    		((IFluidContainerItem)aStack.getItem()).fill(aStack = Stacks.copyAmount(1, aStack), aFluid, true);
				return aStack;
	    	}
	    	Map<Fluid, FluidContainerData> tFluidToContainer = sEmptyContainerToFluidToData.get(new ItemStackContainer(aStack));
	    	if (tFluidToContainer == null) return null;
	    	FluidContainerData tData = tFluidToContainer.get(aFluid.getFluid());
	    	if (tData == null || tData.fluid.amount > aFluid.amount) return null;
	    	if (aRemoveFluidDirectly) aFluid.amount -= tData.fluid.amount;
			return Stacks.copyAmount(1, tData.filledContainer);
	    }
	    
	    public static boolean containsFluid(ItemStack aStack, FluidStack aFluid, boolean aCheckIFluidContainerItems) {
			if (Stacks.invalid(aStack) || aFluid == null) return false;
	    	if (aCheckIFluidContainerItems && aStack.getItem() instanceof IFluidContainerItem && ((IFluidContainerItem)aStack.getItem()).getCapacity(aStack) > 0) return aFluid.isFluidEqual(((IFluidContainerItem)aStack.getItem()).getFluid(aStack = Stacks.copyAmount(1, aStack)));
	    	FluidContainerData tData = sFilledContainerToData.get(new ItemStackContainer(aStack));
	    	return tData==null?false:tData.fluid.isFluidEqual(aFluid);
	    }
	    
		public static FluidStack getFluidForFilledItem(ItemStack aStack, boolean aCheckIFluidContainerItems) {
			if (Stacks.invalid(aStack)) return null;
			if (aCheckIFluidContainerItems && aStack.getItem() instanceof IFluidContainerItem && ((IFluidContainerItem)aStack.getItem()).getCapacity(aStack) > 0) return ((IFluidContainerItem)aStack.getItem()).drain(Stacks.copyAmount(1, aStack), Integer.MAX_VALUE, true);
			FluidContainerData tData = sFilledContainerToData.get(new ItemStackContainer(aStack));
			return tData==null?null:tData.fluid.copy();
		}
		
		public static ItemStack getContainerForFilledItem(ItemStack aStack, boolean aCheckIFluidContainerItems) {
			if (Stacks.invalid(aStack)) return null;
			FluidContainerData tData = sFilledContainerToData.get(new ItemStackContainer(aStack));
			if (tData != null) return Stacks.copyAmount(1, tData.emptyContainer);
			if (aCheckIFluidContainerItems && aStack.getItem() instanceof IFluidContainerItem && ((IFluidContainerItem)aStack.getItem()).getCapacity(aStack) > 0) {
				((IFluidContainerItem)aStack.getItem()).drain(aStack = Stacks.copyAmount(1, aStack), Integer.MAX_VALUE, true);
				return aStack;
			}
			return null;
		}
	}
	
	public static class Books {
		public static final Map<String, ItemStack> BOOK_MAP = new HashMap<String, ItemStack>();
		public static final List<String> BOOK_LIST = new ArrayListNoNulls<String>();
		public static final List<String> MATERIAL_DICTIONARIES = new ArrayListNoNulls<String>();
		
		public static ItemStack getWrittenBook(String aMapping) {
			return getWrittenBook(aMapping, null);
		}
		
		public static ItemStack getWrittenBook(String aMapping, ItemStack aStackToPutNBT) {
			ItemStack tStack = BOOK_MAP.get(aMapping);
			if (tStack == null) return aStackToPutNBT==null?new ItemStack(Items.written_book, 1, 0):aStackToPutNBT;
			if (aStackToPutNBT == null) aStackToPutNBT = Stacks.copy(tStack);
			aStackToPutNBT.setTagCompound(tStack.getTagCompound());
			return aStackToPutNBT;
		}
		
		public static ItemStack createWrittenBook(String aMapping, String aTitle, String aAuthor, ItemStack aDefaultBook, String... aPages) {
			return createWrittenBook(aMapping, aTitle, aAuthor, aDefaultBook, true, aPages);
		}
		
		public static ItemStack createWrittenBook(String aMapping, String aTitle, String aAuthor, ItemStack aDefaultBook, boolean aLogging, String... aPages) {
			if (Code.stringInvalid(aMapping)) return null;
			ItemStack rStack = BOOK_MAP.get(aMapping);
			if (rStack == null) rStack = aDefaultBook==null?new ItemStack(Items.written_book, 1, 0):Stacks.copyAmount(1, aDefaultBook);
			NBTTagCompound rNBT = rStack.getTagCompound();
			if (rNBT != null) return Stacks.copy(rStack);
			if (Code.stringInvalid(aTitle) || Code.stringInvalid(aAuthor) || aPages.length <= 0) return null;
			rNBT = new NBTTagCompound();
	        rNBT.setString("title", aTitle);
	        rNBT.setString("author", aAuthor);
	        NBTTagList tNBTList = new NBTTagList();
	        for (byte i = 0; i < aPages.length; i++) {
		        if (i == 49 && D1) GT_Log.out.println("NOTE: Written Book has more than 50 Pages -> " + aMapping);
		        if (aPages[i].length() < 256)
		        	tNBTList.appendTag(new NBTTagString(aPages[i]));
		        else
		        	GT_Log.err.println("WARNING: String for Page of written Book too long! -> " + aPages[i]);
	        }
	        rNBT.setTag("pages", tNBTList);
	        if (aLogging) GT_Log.out.println("GT_Mod: Added Book to Book List  -  Mapping: '"+aMapping+"'  -  Name: '"+aTitle+"'  -  Author: '"+aAuthor+"'");
	        rStack.setTagCompound(rNBT);
	        BOOK_MAP.put(aMapping, Stacks.copy(rStack));
	        BOOK_LIST.add(aMapping);
			return Stacks.copy(rStack);
		}
		
		public static ItemStack createMaterialDictionary(OreDictMaterial aMat, ItemStack aDefaultBook) {
	        String tPage = "";
	        List<String> tBook = new ArrayListNoNulls<String>();
	        boolean temp = false;
	        int tCounter = 0;
			
        	tBook.add("===================\n"+aMat.getLocal()+"\n===================\nID: "+(aMat.mID<0?"NONE":aMat.mID)+"\nMelting: "+aMat.mMeltingPoint+" K\nBoiling: "+aMat.mBoilingPoint+" K\nPlasma: "+aMat.mPlasmaPoint+" K\n===================\nDensity:\n"+aMat.mGramPerCubicCentimeter+" g/cm3\n"+aMat.getWeight(U)+" kg/unit\n===================\n");
        	
        	//----------
        	
        	if (aMat.mComponents == null) {
        		if (aMat.contains(TD.Atomic.ELEMENT)) {
            		temp = true;
        			tPage="Atomic Structure:\nProtons: "+aMat.mProtons+"\nElectrons: " + aMat.mElectrons + "\nNeutrons: " + aMat.mNeutrons + "\nMass: "+aMat.mMass+"\n===================\n";
        			for (TagData tTag : TD.Atomic.ALL) if (tTag != TD.Atomic.ELEMENT && tTag != TD.Atomic.NONMETAL && aMat.contains(tTag)) tPage += tTag.getLocalisedNameLong() + "\n";
        		}
        	} else {
        		temp = true;
        		tPage="Components per "+aMat.mComponents.getCommonDivider() + "\n===================\n";
        		for (OreDictMaterialStack tMaterial : aMat.mComponents.getComponents()) tPage += ((tMaterial.mAmount * aMat.mComponents.getCommonDivider()) / U)+" "+tMaterial.mMaterial.getLocal()+"\n";
        	}
        	
        	if (temp) {tBook.add(tPage); temp = false;}
        	
        	//----------
        	
    		tPage="Tool Properties\n===================\n";
    		
    		if (aMat.contains(TD.Properties.HAS_TOOL_STATS)) {
    			temp = true; 
    			tPage+="Durability:\n"+aMat.mToolDurability+"\nQuality:\n"+aMat.mToolQuality+"\nSpeed:\n"+aMat.mToolSpeed+"\nHandle:\n"+aMat.mHandleMaterial.getLocal()+"\n";
    		}
    		
        	if (aMat.mEnchantmentTools != null || aMat.mEnchantmentArmors != null) {
        		temp = true; 
            	if (aMat.mEnchantmentTools  != null) tPage += "Tool Enchantment:\n"  + aMat.mEnchantmentTools .getTranslatedName(aMat.mEnchantmentToolsLevel ) + "\n";
            	if (aMat.mEnchantmentArmors != null) tPage += "Armor Enchantment:\n" + aMat.mEnchantmentArmors.getTranslatedName(aMat.mEnchantmentArmorsLevel) + "\n";
        	}
        	
        	if (temp) {tBook.add(tPage+"===================\n"); temp = false;}
        	
        	//----------
        	
    		tPage="Properties\n===================\n";
    		
    		for (TagData tTag : TD.Properties.ALL_RELEVANTS) if (aMat.contains(tTag)) {temp = true; tPage += tTag.getLocalisedNameLong() + "\n";}
    		
        	if (temp) {tBook.add(tPage); temp = false;}
        	
        	//----------
        	
    		tPage="Machine Processing\n===================\n";
    		
    		for (TagData tTag : TD.Processing.ALL_MACHINES) if (aMat.contains(tTag)) {temp = true; tPage += tTag.getLocalisedNameLong() + "\n";}
    		
        	if (temp) {tBook.add(tPage); temp = false;}
        	
        	//----------
        	
    		tPage="Materials which can be decomposed to this\n===================\n";
    		tCounter = 0;
        	for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mComponents != null && tMat.contains(TD.Compounds.DECOMPOSABLE)) {
        		for (OreDictMaterialStack tMt2 : tMat.mComponents.getComponents()) if (tMt2.mMaterial == aMat) {
                	temp=!(tCounter++%6==5);
                	tPage += tMat.getLocal()+"\n";
                    if (!temp) {
                		tBook.add(tPage);
                		tPage="Materials which can be decomposed to this\n===================\n";
                	}
        			break;
        		}
        	}
        	
            if (temp) {tBook.add(tPage); temp = false;}
        	
        	//----------
        	
    		tPage="Ore Processing\n===================\n";
    		
    		for (TagData tTag : TD.Processing.ALL_ORES) if (aMat.contains(tTag)) {temp = true; tPage += tTag.getLocalisedNameLong() + "\n";}
    		
        	if (temp) {tBook.add(tPage); temp = false;}
        	
        	//----------
        	
    		tPage="Ore Byproducts\n===================\n";
    		
    		for (OreDictMaterial tMat : aMat.mByProducts) {temp = true; tPage += tMat.getLocal() + "\n";}
    		
        	if (temp) {tBook.add(tPage); temp = false;}
        	
        	//----------
        	
    		tPage="Ores with this as Byproduct\n===================\n";
    		tCounter = 0;
        	for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mByProducts.contains(aMat)) {
            	temp=!(tCounter++%6==5);
            	tPage += tMat.getLocal()+"\n";
                if (!temp) {
            		tBook.add(tPage);
            		tPage="Ores with this as Byproduct\n===================\n";
            	}
        	}
        	
            if (temp) {tBook.add(tPage); temp = false;}
        	
        	//----------
        	
        	tPage = "Processing Data\n===================\n";
        	tPage += "Smelting:\n"		+(aMat.mTargetSmelting		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetSmelting		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetSmelting	.mMaterial.getLocal()+"\n";
        	tPage += "Burning:\n"		+(aMat.mTargetBurning		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetBurning		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetBurning		.mMaterial.getLocal()+"\n";
        	tPage += "Pulverising:\n"	+(aMat.mTargetPulver		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetPulver		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetPulver		.mMaterial.getLocal()+"\n";
        	tPage += "Bending:\n"		+(aMat.mTargetBending		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetBending		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetBending		.mMaterial.getLocal()+"\n";
        	tPage += "Compressing:\n"	+(aMat.mTargetCompressing	.mAmount / U) + "." + ((int)(((double)(aMat.mTargetCompressing	.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetCompressing	.mMaterial.getLocal()+"\n";
        	
        	tBook.add(tPage);
        	
        	//----------
        	
        	tPage = "Processing Data\n===================\n";
        	tPage += "Crushing:\n"		+(aMat.mTargetCrushing		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetCrushing		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetCrushing	.mMaterial.getLocal()+"\n";
        	tPage += "Cutting:\n"		+(aMat.mTargetCutting		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetCutting		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetCutting		.mMaterial.getLocal()+"\n";
        	tPage += "Forging:\n"		+(aMat.mTargetForging		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetForging		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetForging		.mMaterial.getLocal()+"\n";
        	tPage += "Smashing:\n"		+(aMat.mTargetSmashing		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetSmashing		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetSmashing	.mMaterial.getLocal()+"\n";
        	tPage += "Working:\n"		+(aMat.mTargetWorking		.mAmount / U) + "." + ((int)(((double)(aMat.mTargetWorking		.mAmount % U) / (double)U) * 1000))+" "+aMat.mTargetWorking		.mMaterial.getLocal()+"\n";
        	
        	tBook.add(tPage);
        	
        	//----------
        	
        	tPage = "Thaumaturgic Data\n===================\nAspects:\n";
        	
        	if (aMat.mAspects.isEmpty()) {
        		tPage += "None\n";
        	} else {
        		for (TC_AspectStack tAspect : aMat.mAspects) tPage += tAspect.mAmount + "x " + tAspect.mAspect.mName + "\n";
        	}
        	
    		tPage += "===================\n";
        	
        	tBook.add(tPage);
        	
        	//----------
        	
        	Map<OreDictMaterial, Long> tMap;
        	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetSmelting	.mMaterial == aMat && tMat.mTargetSmelting		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetSmelting		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to smelt for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to smelt for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetBurning		.mMaterial == aMat && tMat.mTargetBurning		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetBurning		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to burn for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to burn for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
         	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetPulver		.mMaterial == aMat && tMat.mTargetPulver		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetPulver		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to pulverise for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to pulverise for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetBending		.mMaterial == aMat && tMat.mTargetBending		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetBending		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to bend for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to bend for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetCompressing	.mMaterial == aMat && tMat.mTargetCompressing	.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetCompressing	.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to compress for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to compress for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetCrushing	.mMaterial == aMat && tMat.mTargetCrushing		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetCrushing		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to crush for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to crush for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetCutting		.mMaterial == aMat && tMat.mTargetCutting		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetCutting		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to cut for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to cut for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetForging		.mMaterial == aMat && tMat.mTargetForging		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetForging		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to forge for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to forge for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetSmashing	.mMaterial == aMat && tMat.mTargetSmashing		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetSmashing		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to smash for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to smash for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	tMap = new HashMap<OreDictMaterial, Long>(); for (OreDictMaterial tMat : OreDictMaterial.MATERIAL_MAP.values()) if (tMat.mTargetRegistration == tMat && tMat != aMat && tMat.mTargetWorking		.mMaterial == aMat && tMat.mTargetWorking		.mAmount > 0 && !tMat.getLocal().equals(aMat.getLocal())) tMap.put(tMat, tMat.mTargetWorking		.mAmount);
        	tMap = UT.Code.sortByValuesDescending(tMap);
        	tPage = "Resources to use in other ways for getting "+aMat.getLocal()+"\n===================\n";
        	tCounter = 0;
        	for (Entry<OreDictMaterial, Long> tEntry : tMap.entrySet()) {
        		temp=!(tCounter++%6==5);
        		tPage+=(tEntry.getValue() / U) + "." + ((int)(((double)(tEntry.getValue() % U) / (double)U) * 1000))+" from 1 "+tEntry.getKey().getLocal()+"\n";
            	if (!temp) {
        			tBook.add(tPage);
                	tPage = "Resources to use in other ways for getting "+aMat.getLocal()+"\n===================\n";
        		}
        	}
        	
        	if (temp) {tBook.add(tPage); temp=false;}
        	
         	//----------
         	
        	for (OreDictMaterial tMat : aMat.mAlloyComponentReferences) {
            	if (tMat.mComponents != null) {
            		tPage="Alloy:\n"+tMat.getLocal()+"\n===================\nMelting: "+tMat.mMeltingPoint+" K\nBoiling: "+tMat.mBoilingPoint+" K\n===================\nComponents per "+tMat.mComponents.getCommonDivider() + "\n";
            		for (OreDictMaterialStack tMt2 : tMat.mComponents.getComponents()) tPage += ((tMt2.mAmount * tMat.mComponents.getCommonDivider()) / U)+" "+tMt2.mMaterial.getLocal()+"\n";
            	 	tBook.add(tPage);
            	}
            }
        	
        	//----------
        	
        	if (aMat.mDescription != null) for (int i = 0, j = 0; i < aMat.mDescription.length; i++) if (Code.stringValid(aMat.mDescription[i])) tBook.add("Description Pg "+(++j)+"\n===================\n" + aMat.mDescription[i]);
        	
        	//----------
        	
        	MATERIAL_DICTIONARIES.add("Material_Dictionary_"+aMat.mNameInternal);
        	
        	return Stacks.copy(aMat.mDictionaryBook = createWrittenBook("Material_Dictionary_"+aMat.mNameInternal, aMat.getLocal()+", Data as of " + DATE_OF_GAME_START, "Material Dictionary Foundation", aDefaultBook==null&&ItemsGT.sMultiItemBooks!=null?new ItemStack(ItemsGT.sMultiItemBooks, 1, tBook.size()>50?32003:32002):aDefaultBook, false, tBook.toArray(new String[tBook.size()])));
		}
	}
	
	public static class Stacks {
		public static boolean debugItem(ItemStack aStack) {
			return /*ItemList.Armor_Cheat.isStackEqual(aStack, T, T) || */equal(GT_ModHandler.getIC2Item("debug", 1), aStack, T);
		}
		
		public static ItemStack update(ItemStack aStack) {
			if (valid(aStack) && aStack.getItem() instanceof IItemUpdatable) ((IItemUpdatable)aStack.getItem()).updateItemStack(aStack);
			return aStack;
		}
		
		public static boolean inList(Collection<ItemStack> aList, ItemStack aStack, boolean aTrueIfListEmpty, boolean aInvertFilter) {
			if (aStack == null || aStack.stackSize < 1) return F;
			if (aList == null) return aTrueIfListEmpty;
			while (aList.contains(null)) aList.remove(null);
			if (aList.size() < 1) return aTrueIfListEmpty;
			Iterator<ItemStack> tIterator = aList.iterator();
			ItemStack tStack = null;
			while (tIterator.hasNext()) if ((tStack = tIterator.next())!= null && equal(aStack, tStack)) return !aInvertFilter;
			return aInvertFilter;
		}
		
		public static ItemStack set(Object aSetStack, Object aToStack) {
			if (invalid(aSetStack) || invalid(aToStack)) return null;
			((ItemStack)aSetStack).func_150996_a(((ItemStack)aToStack).getItem());
			((ItemStack)aSetStack).stackSize = ((ItemStack)aToStack).stackSize;
			Items.feather.setDamage((ItemStack)aSetStack, Items.feather.getDamage((ItemStack)aToStack));
			((ItemStack)aSetStack).setTagCompound(((ItemStack)aToStack).getTagCompound());
			return (ItemStack)aSetStack;
		}
		
		public static ItemStack container(ItemStack aStack, boolean aCheckIFluidContainerItems) {
			if (invalid(aStack)) return null;
			if (aStack.getItem().hasContainerItem(aStack)) return aStack.getItem().getContainerItem(aStack);
			/** These are all special Cases, in which it is intended to have only GT Blocks outputting those Container Items */
			if (IL.Cell_Empty.equal(aStack, false, true)) return null;
			if (aStack.getItem() == Items.potionitem || aStack.getItem() == Items.experience_bottle || IL.TF_Vial_FieryBlood.equal(aStack) || IL.TF_Vial_FieryTears.equal(aStack)) return IL.Bottle_Empty.get(1);
			
			if (aCheckIFluidContainerItems && aStack.getItem() instanceof IFluidContainerItem && ((IFluidContainerItem)aStack.getItem()).getCapacity(aStack) > 0) {
				ItemStack tStack = copyAmount(1, aStack);
				((IFluidContainerItem)aStack.getItem()).drain(tStack, Integer.MAX_VALUE, true);
				if (!equal(aStack, tStack)) return tStack;
				return null;
			}
			
			if (IL.IC2_ForgeHammer.equal(aStack) || IL.IC2_WireCutter.equal(aStack)) return copyMeta(Items.feather.getDamage(aStack) + 1, aStack);
			return null;
		}
		
		public static boolean equal(ItemStack aStack1, ItemStack aStack2) {
			return equal(aStack1, aStack2, false);
		}
		
		public static boolean equal(ItemStack aStack1, ItemStack aStack2, boolean aIgnoreNBT) {
			return aStack1 != null && aStack2 != null && aStack1.getItem() == aStack2.getItem() && (aIgnoreNBT || ((aStack1.getTagCompound() == null) == (aStack2.getTagCompound() == null)) && (aStack1.getTagCompound() == null || aStack1.getTagCompound().equals(aStack2.getTagCompound()))) && (Items.feather.getDamage(aStack1) == Items.feather.getDamage(aStack2) || Items.feather.getDamage(aStack1) == CS.W || Items.feather.getDamage(aStack2) == CS.W);
		}
		
		public static boolean equalUnification(ItemStack aStack1, ItemStack aStack2) {
			return equalUnification(aStack1, aStack2, false);
		}
		
		public static boolean equalUnification(ItemStack aStack1, ItemStack aStack2, boolean aIgnoreNBT) {
			return equal(OM.get(aStack1), OM.get(aStack2), aIgnoreNBT);
		}
		
		public static ItemStack create(long aItemID, long aStacksize, long aMetaData, NBTTagCompound aNBT) {
			return create(Item.getItemById((int)aItemID), aStacksize, aMetaData, aNBT);
		}
		
		public static ItemStack create(Item aItem, long aStacksize, long aMetaData, NBTTagCompound aNBT) {
			return aItem == null ? null : create(new ItemStack(aItem, (byte)aStacksize, (int)aMetaData), aNBT);
		}
		
		public static ItemStack create(Block aBlock, long aStacksize, long aMetaData, NBTTagCompound aNBT) {
			return aBlock == null ? null : create(new ItemStack(aBlock, (byte)aStacksize, (int)aMetaData), aNBT);
		}
		
		public static ItemStack create(ItemStack aStack, NBTTagCompound aNBT) {
			if (aStack == null) return null;
			aStack = aStack.copy();
			aStack.setTagCompound(aNBT);
			return aStack;
		}
		
		public static ItemStack[] copyArray(Object... aStacks) {
			ItemStack[] rStacks = new ItemStack[aStacks.length];
			for (int i = 0; i < aStacks.length; i++) rStacks[i] = copy(aStacks[i]);
			return rStacks;
		}
		
		public static ItemStack copy(Object... aStacks) {
			for (Object tStack : aStacks) if (valid(tStack)) return ((ItemStack)tStack).copy();
			return null;
		}
		
		public static ItemStack copyAmount(long aAmount, Object... aStacks) {
			ItemStack rStack = copy(aStacks);
			if (invalid(rStack)) return null;
			if (aAmount > 64) aAmount = 64; else if (aAmount == -1) aAmount = 111; else if (aAmount < 0) aAmount = 0;
			rStack.stackSize = (byte)aAmount;
			return rStack;
		}
		
		public static ItemStack copyMeta(long aMetaData, Object... aStacks) {
			ItemStack rStack = copy(aStacks);
			if (invalid(rStack)) return null;
			Items.feather.setDamage(rStack, (short)aMetaData);
			return rStack;
		}
		
		public static ItemStack copyAmountAndMeta(long aAmount, long aMetaData, Object... aStacks) {
			ItemStack rStack = copyAmount(aAmount, aStacks);
			if (invalid(rStack)) return null;
			Items.feather.setDamage(rStack, (short)aMetaData);
			return rStack;
		}
		
		/**
		 * returns a copy of an ItemStack with its Stacksize being multiplied by aMultiplier
		 */
		public static ItemStack mul(long aMultiplier, Object... aStacks) {
			ItemStack rStack = copy(aStacks);
			if (rStack == null) return null;
			rStack.stackSize *= aMultiplier;
			return rStack;
		}
		
		public static boolean inList(ItemStack aStack, Collection<ItemStackContainer> aList) {
			if (aStack == null) return false;
			return inList(new ItemStackContainer(aStack), aList);
		}
		
		public static boolean inList(ItemStackContainer aStack, Collection<ItemStackContainer> aList) {
			return aStack != null && (aList.contains(aStack) || aList.contains(new ItemStackContainer(aStack.mItem, aStack.mStackSize, W)));
		}
		
		public static int toInt(ItemStack aStack) {
			if (invalid(aStack)) return 0;
			return Item.getIdFromItem(aStack.getItem()) | (Items.feather.getDamage(aStack)<<16);
		}
		
		public static int toIntWildcard(ItemStack aStack) {
			if (invalid(aStack)) return 0;
			return Item.getIdFromItem(aStack.getItem()) | (W<<16);
		}
		
		public static ItemStack toStack(int aStack) {
		    Item tItem = Item.getItemById(aStack&(~0>>>16));
			if (tItem != null) return new ItemStack(tItem, 1, aStack>>>16);
			return null;
		}
		
		public static Integer[] toIntegerArray(ItemStack... aStacks) {
			Integer[] rArray = new Integer[aStacks.length];
			for (int i = 0; i < rArray.length; i++) rArray[i] = toInt(aStacks[i]);
			return rArray;
		}
		
		public static int[] toIntArray(ItemStack... aStacks) {
			int[] rArray = new int[aStacks.length];
			for (int i = 0; i < rArray.length; i++) rArray[i] = toInt(aStacks[i]);
			return rArray;
		}
		
		public static Block block(Object aStack) {
			if (invalid(aStack)) return Blocks.air;
			return Block.getBlockFromItem(((ItemStack)aStack).getItem());
		}
		
		public static boolean valid(Object aStack) {
			return aStack != null &&  (aStack instanceof ItemStack) && ((ItemStack)aStack).getItem() != null && ((ItemStack)aStack).stackSize >= 0;
		}
		
		public static boolean invalid(Object aStack) {
			return aStack == null || !(aStack instanceof ItemStack) || ((ItemStack)aStack).getItem() == null || ((ItemStack)aStack).stackSize <  0;
		}
		
		public static void hide(Item aItem) {
			hide(aItem, W);
		}
		
		public static void hide(Item aItem, long aMetaData) {
			hide(new ItemStack(aItem, 1, (short)aMetaData));
		}
		
		public static void hide(Block aBlock) {
			hide(aBlock, W);
		}
		
		public static void hide(Block aBlock, long aMetaData) {
			hide(new ItemStack(aBlock, 1, (short)aMetaData));
		}
		
		public static void hide(ItemStack aStack) {
			try {codechicken.nei.api.API.hideItem(new ItemStack(aStack.getItem(), 1, Items.feather.getDamage(aStack)));} catch(Throwable e) {/**/}
		}
		
		/**
		 * Loads an ItemStack properly.
		 */
		public static ItemStack load(NBTTagCompound aNBT, String aTagName) {
			return load(aNBT.getCompoundTag(aTagName));
		}
		
		/**
		 * Loads an ItemStack properly.
		 */
		public static ItemStack load(NBTTagCompound aNBT) {
			if (aNBT == null) return null;
			ItemStack rStack = ItemStack.loadItemStackFromNBT(aNBT);
			try {
				if (rStack != null && (rStack.getItem().getClass().getName().startsWith("ic2.core.migration"))) rStack.getItem().onUpdate(rStack, CS.DW, null, 0, false);
			} catch(Throwable e) {
				FMLLog.warning("%s", e.toString());
			}
			return OM.get(rStack);
		}
		
		/**
		 * Saves an ItemStack properly.
		 */
		public static NBTTagCompound save(NBTTagCompound aNBT, String aTagName, ItemStack aStack) {
			NBTTagCompound rNBT = save(aStack);
			if (rNBT != null) aNBT.setTag(aTagName, rNBT);
			return rNBT;
		}
		
		/**
		 * Saves an ItemStack properly.
		 */
		public static NBTTagCompound save(ItemStack aStack) {
			return aStack == null ? null : OM.get(aStack).writeToNBT(new NBTTagCompound());
		}
	}
	
	public static class Code {
		/** @return an unsigned representation of this Byte. */
		public static short unsignB(byte aByte) {
			return aByte < 0 ? (short)(aByte + 256) : aByte;
		}
		
		/** @return an unsigned representation of this Short. */
		public static int unsignS(short aShort) {
			return aShort < 0 ? aShort + 65536 : aShort;
		}
		
		/** @return an unsigned representation of this Integer. */
		public static long unsignI(int aInteger) {
			return aInteger < 0 ? aInteger + 4294967296L : aInteger;
		}
		
		public static byte toByteS(short aValue, int aIndex) {return (byte)(0xff & (aValue >> (aIndex<<3)));}
		public static byte toByteI(int   aValue, int aIndex) {return (byte)(0xff & (aValue >> (aIndex<<3)));}
		public static byte toByteL(long  aValue, int aIndex) {return (byte)(0xff & (aValue >> (aIndex<<3)));}
		
		public static short combine(byte aValue1, byte aValue2)																						{return (short)((0xff & aValue1) | (0xff & aValue2) << 8);}
		public static int   combine(byte aValue1, byte aValue2, byte aValue3, byte aValue4)															{return         (0xff & aValue1) | (0xff & aValue2) << 8 | (0xff & aValue3) << 16 | (0xff & aValue4) << 24;}
		public static long  combine(byte aValue1, byte aValue2, byte aValue3, byte aValue4, byte aValue5, byte aValue6, byte aValue7, byte aValue8)	{return         (0xff & aValue1) | (0xff & aValue2) << 8 | (0xff & aValue3) << 16 | (0xff & aValue4) << 24 | (0xff & aValue5) << 32 | (0xff & aValue6) << 40 | (0xff & aValue7) << 48 | (0xff & aValue8) << 56;}
		
		public static long getBits(boolean... aBits) {
			long rBits = 0;
			for (int i = 0; i < 64 && i < aBits.length; i++) if (aBits[i]) rBits |= (1 << i);
			return rBits;
		}
		
		public static boolean[] getBitsB(byte aBits) {
			boolean[] rBits = new boolean[8];
			for (int i = 0; i < rBits.length; i++) if ((aBits & (1 << i)) != 0) rBits[i] = true;
			return rBits;
		}
		
		public static boolean[] getBitsS(short aBits) {
			boolean[] rBits = new boolean[16];
			for (int i = 0; i < rBits.length; i++) if ((aBits & (1 << i)) != 0) rBits[i] = true;
			return rBits;
		}
		
		public static boolean[] getBitsI(int aBits) {
			boolean[] rBits = new boolean[32];
			for (int i = 0; i < rBits.length; i++) if ((aBits & (1 << i)) != 0) rBits[i] = true;
			return rBits;
		}
		
		public static boolean[] getBitsL(long aBits) {
			boolean[] rBits = new boolean[64];
			for (int i = 0; i < rBits.length; i++) if ((aBits & (1 << i)) != 0) rBits[i] = true;
			return rBits;
		}
		
		public static ItemStack toStack(int aStack) {
		    Item tItem = Item.getItemById(aStack&(~0>>>16));
			if (tItem != null) return new ItemStack(tItem, 1, aStack>>>16);
			return null;
		}
		
		/** Converts a Number to a String */
	    public static String makeString(long aNumber) {
	    	String tString = "";
	    	boolean temp = T, negative = F;
	    	
	    	if (aNumber<0) {
	    		aNumber *= -1;
	    		negative = T;
	    	}
	    	
	    	for (long i = 1000000000000000000L; i > 0; i /= 10) {
	    		long tDigit = (aNumber/i)%10;
				if ( temp && tDigit != 0) temp = F;
				if (!temp) {
					tString += tDigit;
					if (i != 1) for (long j = i; j > 0; j /= 1000) if (j == 1) tString += ",";
				}
	    	}
	    	
	    	if (tString.equals("")) tString = "0";
	    	
	    	return negative?"-"+tString:tString;
	    }
	    
		public static boolean containsBoolean(boolean aTarget, boolean... aArray) {
			if (aArray != null) for (boolean tValue : aArray) if (tValue == aTarget) return T;
			return F;
		}
		
		public static <E> boolean containsSomething(E... aArray) {
			if (aArray != null) for (Object tObject : aArray) if (tObject != null) return T;
			return F;
		}
		
		public static <E> ArrayList<E> getWithoutNulls(E... aArray) {
			if (aArray == null) return new ArrayListNoNulls<E>();
			ArrayList<E> rList = new ArrayListNoNulls<E>(Arrays.asList(aArray));
			return rList;
		}
		
		public static <E> ArrayList<E> getWithoutTrailingNulls(E... aArray) {
			if (aArray == null) return new ArrayList<E>(1);
			ArrayList<E> rList = new ArrayList<E>(Arrays.asList(aArray));
			for (int i = rList.size() - 1; i >= 0 && rList.get(i) == null;) rList.remove(i--);
			return rList;
		}
		
		private static final DateFormat sDateFormat = DateFormat.getInstance();
	    public static String dateAndTime() {
	    	return sDateFormat.format(new Date());
	    }
	    
		public static byte tier(long l) {
			byte i = -1;
			while (++i < V.length) if (l <= V[i]) return i;
			return i;
		}
		
	    /** re-maps all Keys of a Map after the Keys were weakened. */
		public static <X, Y> Map<X, Y> reMap(Map<X, Y> aMap) {
			Map<X, Y> tMap = new HashMap<X, Y>();
			tMap.putAll(aMap);
			aMap.clear();
			aMap.putAll(tMap);
			return aMap;
		}
		
		/** Why the fuck do neither Java nor Guava have a Function to do this? */
	    public static <X, Y extends Comparable> LinkedHashMap<X,Y> sortByValuesAcending(Map<X,Y> aMap) {
	        List<Map.Entry<X,Y>> tEntrySet = new LinkedList<Map.Entry<X,Y>>(aMap.entrySet());
	        Collections.sort(tEntrySet, new Comparator<Map.Entry<X,Y>>() {@Override public int compare(Entry<X, Y> aValue1, Entry<X, Y> aValue2) {return aValue1.getValue().compareTo(aValue2.getValue());}});
	        LinkedHashMap<X,Y> rMap = new LinkedHashMap<X,Y>();
	        for (Map.Entry<X,Y> tEntry : tEntrySet) rMap.put(tEntry.getKey(), tEntry.getValue());
	        return rMap;
	    }
	    
		/** Why the fuck do neither Java nor Guava have a Function to do this? */
	    public static <X, Y extends Comparable> LinkedHashMap<X,Y> sortByValuesDescending(Map<X,Y> aMap) {
	        List<Map.Entry<X,Y>> tEntrySet = new LinkedList<Map.Entry<X,Y>>(aMap.entrySet());
	        Collections.sort(tEntrySet, new Comparator<Map.Entry<X,Y>>() {@Override public int compare(Entry<X, Y> aValue1, Entry<X, Y> aValue2) {return -aValue1.getValue().compareTo(aValue2.getValue());}});
	        LinkedHashMap<X,Y> rMap = new LinkedHashMap<X,Y>();
	        for (Map.Entry<X,Y> tEntry : tEntrySet) rMap.put(tEntry.getKey(), tEntry.getValue());
	        return rMap;
	    }
	    
		public static <E> E selectInList(int aIndex, E aReplacement, List<E> aList) {
			if (aList == null || aList.isEmpty()) return aReplacement;
			if (aList.size() <= aIndex) return aList.get(aList.size() - 1);
			if (aIndex < 0) return aList.get(0);
			return aList.get(aIndex);
		}
		
		public static <E> E selectInArray(int aIndex, E aReplacement, E... aList) {
			if (aList == null || aList.length == 0) return aReplacement;
			if (aList.length <= aIndex) return aList[aList.length - 1];
			if (aIndex < 0) return aList[0];
			return aList[aIndex];
		}
		
		public static boolean inArray(Object aObject, Object... aObjects) {
			return inList(aObject, Arrays.asList(aObjects));
		}
		
		public static boolean inList(Object aObject, Collection aObjects) {
			if (aObjects == null) return false;
			return aObjects.contains(aObject);
		}
		
		public static boolean stringValid(Object aString) {
			return aString != null && !aString.toString().isEmpty();
		}
		
		public static boolean stringInvalid(Object aString) {
			return aString == null || aString.toString().isEmpty();
		}
		
		public static byte side(ForgeDirection aDirection) {
			return (byte)(aDirection==null?SIDE_INVALID:aDirection.ordinal());
		}
		
		/** If this Index exists inside the passed Array and if it is != null */
		public static <E> boolean exists(int aIndex, E[] aArray) {
			return aIndex >= 0 && aIndex < aArray.length && aArray[aIndex] != null;
		}
		
		/** @return a Value for a Scale between 0 and aMax with aScale+1 possible Steps. 0 is only returned if the aValue is <= 0, aScale is only returned if the Value is >= aMax. The remaining values between ]0:aScale[ are returned for each Step of the Scale. This Function finds use in Displays such as the Barometer, but also in Redstone. */
		public static long scale(long aValue, long aMax, long aScale, boolean aInvert) {
			long rScale = (aValue <= 0 ? 0 : aValue >= aMax ? aScale : aScale <= 2 ? 1 : 1 + (aValue * (aScale-1)) / aMax);
			return aInvert ? aScale - rScale : rScale;
		}
		
		/** @return a Value for a Scale between aMin and aMax with aScale+1 possible Steps. 0 is only returned if the aValue is <= aMin, aScale is only returned if the Value is >= aMax. The remaining values between ]0:aScale[ are returned for each Step of the Scale. This Function finds use in Displays such as the Barometer, but also in Redstone. */
		public static long scale(long aValue, long aMin, long aMax, long aScale, boolean aInvert) {
			return scale(aValue-aMin, aMax-aMin, aScale, aInvert);
		}
		
		public static long bind(long aMin, long aMax, long aBoundValue) {
			return aMin > aMax ? Math.max(aMax, Math.min(aMin, aBoundValue)) : Math.max(aMin, Math.min(aMax, aBoundValue));
		}
		
		public static int bindInt(long aBoundValue) {
			return (int)bind(Integer.MIN_VALUE, Integer.MAX_VALUE, aBoundValue);
		}
		
		public static short bindByte(long aColor) {
			return (short)Math.max(0, Math.min(255, aColor));
		}
		
		public static short[] bindRGBa(short[] aColors) {
			if (aColors == null) return new short[] {255,255,255,255};
			for (int i = 0; i < aColors.length; i++) aColors[i] = bindByte(aColors[i]);
			return aColors;
		}
		
		public static int mixRGBInt(int aRGB1, int aRGB2) {
			return getRGBInt(new short[] {(short)((UT.Code.getR(aRGB1) + UT.Code.getR(aRGB2)) >> 1), (short)((UT.Code.getG(aRGB1) + UT.Code.getG(aRGB2)) >> 1), (short)((UT.Code.getB(aRGB1) + UT.Code.getB(aRGB2)) >> 1)});
		}
		
		public static int getRGBInt(short[] aColors) {
			return aColors == null ? 16777215 : (bindByte(aColors[0]) << 16) | (bindByte(aColors[1]) << 8) | bindByte(aColors[2]);
		}
		
		public static int getRGBaInt(short[] aColors) {
			return aColors == null ? 16777215 : (bindByte(aColors[0]) << 16) | (bindByte(aColors[1]) << 8) | bindByte(aColors[2]) | (bindByte(aColors[3]) << 24);
		}
		
		public static short[] getRGBaArray(int aColors) {
			return new short[] {(short)((aColors >> 16) & 255), (short)((aColors >> 8) & 255), (short)(aColors & 255), (short)((aColors >> 24) & 255)};
		}
		
		public static short getR(int aColors) {return (short)((aColors >> 16) & 255);}
		public static short getG(int aColors) {return (short)((aColors >> 8) & 255);}
		public static short getB(int aColors) {return (short)(aColors & 255);}
		public static short getA(int aColors) {return (short)((aColors >> 24) & 255);}
		
		/** toUpperCases the first Character of the String and returns it */
		public static String capitalise(String aString) {
			if (aString != null && !aString.isEmpty()) return aString.substring(0, 1).toUpperCase() + aString.substring(1);
			return "";
		}
		
		/** @return the opposite facing of this Side of a Block, with a boundary check. */
		public static byte opposite(int aSide) {
			return aSide < OPPOSITES.length && aSide >= 0 ? OPPOSITES[aSide] : 6;
		}
		
	    /** Translates Amount of aUnit1 to Amount of aUnit2. */
	    public static long units(long aAmount, long aUnit1, long aUnit2, boolean aRoundUp) {
	    	if (aUnit1 == aUnit2) return aAmount;
	    	return Math.max(0, ((aAmount * aUnit2) / aUnit1) + (aRoundUp && (aAmount * aUnit2) % aUnit1 > 0 ? 1 : 0));
	    }
	    
		/** @return an Array containing the X and the Y Coordinate of the clicked Point, with the top left Corner as Origin, like on the Texture Sheet. return values should always be between [0.0F and 0.99F]. */
		public static float[] getFacingCoordsClicked(byte aSide, float aHitX, float aHitY, float aHitZ) {
			switch (aSide) {
			case  0: return new float[] {Math.min(0.99F, Math.max(0,1-aHitX)), Math.min(0.99F, Math.max(0,  aHitZ))};
			case  1: return new float[] {Math.min(0.99F, Math.max(0,  aHitX)), Math.min(0.99F, Math.max(0,  aHitZ))};
			case  2: return new float[] {Math.min(0.99F, Math.max(0,1-aHitX)), Math.min(0.99F, Math.max(0,1-aHitY))};
			case  3: return new float[] {Math.min(0.99F, Math.max(0,  aHitX)), Math.min(0.99F, Math.max(0,1-aHitY))};
			case  4: return new float[] {Math.min(0.99F, Math.max(0,  aHitZ)), Math.min(0.99F, Math.max(0,1-aHitY))};
			case  5: return new float[] {Math.min(0.99F, Math.max(0,1-aHitZ)), Math.min(0.99F, Math.max(0,1-aHitY))};
			default: return new float[] {0.5F, 0.5F};
			}
		}
		
		public static byte getSideForPlayerPlacing(Entity aPlayer) {
			if (aPlayer.rotationPitch >=  65) return SIDE_UP;
			if (aPlayer.rotationPitch <= -65) return SIDE_DOWN;
			return COMPASS_DIRECTIONS[MathHelper.floor_double(4*aPlayer.rotationYaw/360+0.5)&3];
		}
		
		public static byte getSideForPlayerPlacing(Entity aPlayer, byte aDefaultFacing, boolean[] aAllowedFacings) {
			if (aPlayer.rotationPitch >=  65 && aAllowedFacings[SIDE_UP]) return SIDE_UP;
			if (aPlayer.rotationPitch <= -65 && aAllowedFacings[SIDE_DOWN]) return SIDE_DOWN;
			byte rFacing = COMPASS_DIRECTIONS[MathHelper.floor_double(4*aPlayer.rotationYaw/360+0.5)&3];
			return aAllowedFacings[rFacing]?rFacing:aDefaultFacing;
		}
		
		/**
		 * This Function determines the direction a Block gets when being Wrenched.
		 * returns -1 if invalid. Even though that could never happen.
		 */
		public static byte getSideWrenching(byte aSide, float aX, float aY, float aZ) {
			switch (aSide) {
			case  0: case  1:
				if (aX < 0.25) return aZ < 0.25 || aZ > 0.75 ? OPPOSITES[aSide] : 4;
				if (aX > 0.75) return aZ < 0.25 || aZ > 0.75 ? OPPOSITES[aSide] : 5;
				if (aZ < 0.25) return 2;
				if (aZ > 0.75) return 3;
				return aSide;
			case  2: case  3:
				if (aX < 0.25) return aY < 0.25 || aY > 0.75 ? OPPOSITES[aSide] : 4;
				if (aX > 0.75) return aY < 0.25 || aY > 0.75 ? OPPOSITES[aSide] : 5;
				if (aY < 0.25) return 0;
				if (aY > 0.75) return 1;
				return aSide;
			case  4: case  5:
				if (aZ < 0.25) return aY < 0.25 || aY > 0.75 ? OPPOSITES[aSide] : 2;
				if (aZ > 0.75) return aY < 0.25 || aY > 0.75 ? OPPOSITES[aSide] : 3;
				if (aY < 0.25) return 0;
				if (aY > 0.75) return 1;
				return aSide;
			}
			return -1;
		}
	}
	
	public static class NBT {
		public static NBTTagCompound getNBTs(NBTTagCompound aNBT, Object... aTags) {
	    	if (aNBT == null) aNBT = new NBTTagCompound();
			for (int i = 1; i < aTags.length; i+=2) {
				     if (aTags[i] instanceof Byte)		aNBT.setByte(		aTags[i-1].toString(), (Byte)		aTags[i]);
				else if (aTags[i] instanceof Short)		aNBT.setShort(		aTags[i-1].toString(), (Short)		aTags[i]);
				else if (aTags[i] instanceof Integer)	aNBT.setInteger(	aTags[i-1].toString(), (Integer)	aTags[i]);
				else if (aTags[i] instanceof Long)		aNBT.setLong(		aTags[i-1].toString(), (Long)		aTags[i]);
				else if (aTags[i] instanceof Float)		aNBT.setFloat(		aTags[i-1].toString(), (Float)		aTags[i]);
				else if (aTags[i] instanceof Double)	aNBT.setDouble(		aTags[i-1].toString(), (Double)		aTags[i]);
				else if (aTags[i] instanceof String)	aNBT.setString(		aTags[i-1].toString(), (String)		aTags[i]);
				else if (aTags[i] instanceof NBTBase)	aNBT.setTag(		aTags[i-1].toString(), (NBTBase)	aTags[i]);
				else 									aNBT.setString(		aTags[i-1].toString(), aTags[i].toString());
			}
	    	return aNBT;
		}
		
		/** Fuses two NBT Compounds together with the Priority lying on the content of the first NBT */
		public static NBTTagCompound fuseNBTs(NBTTagCompound aNBT1, NBTTagCompound aNBT2) {
			if (aNBT1 == null) return aNBT2==null?new NBTTagCompound():(NBTTagCompound)aNBT2.copy();
			NBTTagCompound rNBT = (NBTTagCompound)aNBT1.copy();
			if (aNBT2 == null) return rNBT;
			for (Object tKey : aNBT2.func_150296_c()) if (!rNBT.hasKey(tKey.toString())) rNBT.setTag(tKey.toString(), aNBT2.getTag(tKey.toString()));
	    	return rNBT;
		}
		
	    public static NBTTagCompound getNBTBoolean(NBTTagCompound aNBT, Object aTag, boolean aValue) {
	    	if (aNBT == null) aNBT = new NBTTagCompound();
	    	aNBT.setBoolean(aTag.toString(), aValue);
	    	return aNBT;
	    }
	    
	    public static NBTTagCompound getNBTByte(NBTTagCompound aNBT, Object aTag, byte aValue) {
	    	if (aNBT == null) aNBT = new NBTTagCompound();
	    	aNBT.setByte(aTag.toString(), aValue);
	    	return aNBT;
	    }
	    
	    public static NBTTagCompound getNBTShort(NBTTagCompound aNBT, Object aTag, short aValue) {
	    	if (aNBT == null) aNBT = new NBTTagCompound();
	    	aNBT.setShort(aTag.toString(), aValue);
	    	return aNBT;
	    }
	    
	    public static NBTTagCompound getNBTInteger(NBTTagCompound aNBT, Object aTag, int aValue) {
	    	if (aNBT == null) aNBT = new NBTTagCompound();
	    	aNBT.setInteger(aTag.toString(), aValue);
	    	return aNBT;
	    }
	    
	    public static NBTTagCompound getNBTFloat(NBTTagCompound aNBT, Object aTag, float aValue) {
	    	if (aNBT == null) aNBT = new NBTTagCompound();
	    	aNBT.setFloat(aTag.toString(), aValue);
	    	return aNBT;
	    }
	    
	    public static NBTTagCompound getNBTDouble(NBTTagCompound aNBT, Object aTag, double aValue) {
	    	if (aNBT == null) aNBT = new NBTTagCompound();
	    	aNBT.setDouble(aTag.toString(), aValue);
	    	return aNBT;
	    }
	    
	    public static NBTTagCompound getNBTString(NBTTagCompound aNBT, Object aTag, Object aValue) {
	    	if (aNBT == null) aNBT = new NBTTagCompound();
	    	if (aValue == null) return aNBT;
	    	aNBT.setString(aTag.toString(), aValue.toString());
	    	return aNBT;
	    }
	    
		public static void setNBT(ItemStack aStack, NBTTagCompound aNBT) {
			if (aNBT == null) {aStack.setTagCompound(null); return;}
			ArrayList<String> tTagsToRemove = new ArrayListNoNulls<String>();
			for (Object tKey : aNBT.func_150296_c()) {
				NBTBase tValue = aNBT.getTag((String)tKey);
				if (tValue == null || (tValue instanceof NBTPrimitive && ((NBTPrimitive)tValue).func_150291_c() == 0) || (tValue instanceof NBTTagString && Code.stringInvalid(((NBTTagString)tValue).func_150285_a_()))) tTagsToRemove.add((String)tKey);
			}
			for (Object tKey : tTagsToRemove) aNBT.removeTag((String)tKey);
			aStack.setTagCompound(aNBT.hasNoTags()?null:aNBT);
		}
		
		public static NBTTagCompound getNBT(ItemStack aStack) {
	    	NBTTagCompound rNBT = aStack.getTagCompound();
			return rNBT==null?new NBTTagCompound():rNBT;
		}
		
	    public static void setPunchCardData(ItemStack aStack, String aPunchCardData) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	tNBT.setString("gt.punchcard", aPunchCardData);
	    	setNBT(aStack, tNBT);
	    }
	    public static String getPunchCardData(ItemStack aStack) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	return tNBT.getString("gt.punchcard");
	    }
	    
	    public static void setLighterFuel(ItemStack aStack, long aFuel) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	tNBT.setLong("gt.lighter", aFuel);
	    	setNBT(aStack, tNBT);
	    }
	    public static long getLighterFuel(ItemStack aStack) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	return tNBT.getLong("gt.lighter");
	    }
	    
	    public static void setMapID(ItemStack aStack, short aMapID) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	tNBT.setShort("map_id", aMapID);
	    	setNBT(aStack, tNBT);
	    }
	    public static short getMapID(ItemStack aStack) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	if (!tNBT.hasKey("map_id")) return -1;
	    	return tNBT.getShort("map_id");
	    }
	    
	    public static void setBookTitle(ItemStack aStack, String aTitle) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	tNBT.setString("title", aTitle);
	    	setNBT(aStack, tNBT);
	    }
	    public static String getBookTitle(ItemStack aStack) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	return tNBT.getString("title");
	    }
	    
		public static void setBookAuthor(ItemStack aStack, String aAuthor) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	tNBT.setString("author", aAuthor);
	    	setNBT(aStack, tNBT);
	    }
		public static String getBookAuthor(ItemStack aStack) {
	    	NBTTagCompound tNBT = getNBT(aStack);
	    	return tNBT.getString("author");
	    }
		
		public static void addEnchantment(ItemStack aStack, Enchantment aEnchantment, int aLevel) {
			NBTTagCompound tNBT = getNBT(aStack), tEnchantmentTag;
	        if (!tNBT.hasKey("ench", 9)) tNBT.setTag("ench", new NBTTagList());
	        NBTTagList tList = tNBT.getTagList("ench", 10);
	        
	        boolean temp = CS.T;
	        
	        for (int i = 0; i < tList.tagCount(); i++) {
	        	tEnchantmentTag = tList.getCompoundTagAt(i);
	        	if (tEnchantmentTag.getShort("id") == aEnchantment.effectId) {
	    	        tEnchantmentTag.setShort("id", (short)aEnchantment.effectId);
	    	        tEnchantmentTag.setShort("lvl", (byte)aLevel);
	    	        temp = F;
	    	        break;
	        	}
	        }
	        
	        if (temp) {
		        tEnchantmentTag = new NBTTagCompound();
		        tEnchantmentTag.setShort("id", (short)aEnchantment.effectId);
		        tEnchantmentTag.setShort("lvl", (byte)aLevel);
		        tList.appendTag(tEnchantmentTag);
	        }
	        aStack.setTagCompound(tNBT);
		}
	}
	
	/**
	 * THIS IS BULLSHIT!!! WHY DO I HAVE TO DO THIS SHIT JUST TO HAVE ENCHANTS PROPERLY!?!
	 */
	public static class Enchantments {
	    private static final BullshitIteratorA mBullshitIteratorA = new BullshitIteratorA();
	    private static final BullshitIteratorB mBullshitIteratorB = new BullshitIteratorB();
	    
	    private static void applyBullshit(IBullshit aBullshitModifier, ItemStack aStack) {
	        if (aStack != null) {
	            NBTTagList nbttaglist = aStack.getEnchantmentTagList();
	            if (nbttaglist != null) {
	                for (int i = 0; i < nbttaglist.tagCount(); ++i) {
	    	        	try {
	    	        	    short short1 = nbttaglist.getCompoundTagAt(i).getShort("id");
		                    short short2 = nbttaglist.getCompoundTagAt(i).getShort("lvl");
		                    if (Enchantment.enchantmentsList[short1] != null) aBullshitModifier.calculateModifier(Enchantment.enchantmentsList[short1], short2);
	    	        	} catch(Throwable e) {
	    	        		//
	    	        	}
	                }
	            }
	        }
	    }
	    
	    private static void applyArrayOfBullshit(IBullshit aBullshitModifier, ItemStack[] aStacks) {
	        for (int i = 0; i < aStacks.length; i++) applyBullshit(aBullshitModifier, aStacks[i]);
	    }
	    
	    public static void applyBullshitA(EntityLivingBase aPlayer, Entity aEntity, ItemStack aStack) {
	        mBullshitIteratorA.mPlayer = aPlayer;
	        mBullshitIteratorA.mEntity = aEntity;
	        if (aPlayer != null) applyArrayOfBullshit(mBullshitIteratorA, aPlayer.getLastActiveItems());
	        if (aStack != null) applyBullshit(mBullshitIteratorA, aStack);
	    }
	    
	    public static void applyBullshitB(EntityLivingBase aPlayer, Entity aEntity, ItemStack aStack) {
	        mBullshitIteratorB.mPlayer = aPlayer;
	        mBullshitIteratorB.mEntity = aEntity;
	        if (aPlayer != null) applyArrayOfBullshit(mBullshitIteratorB, aPlayer.getLastActiveItems());
	        if (aStack != null) applyBullshit(mBullshitIteratorB, aStack);
	    }
	    
	    static final class BullshitIteratorA implements IBullshit {
	        public EntityLivingBase mPlayer;
	        public Entity mEntity;
	        BullshitIteratorA() {}
	        
	        @Override
			public void calculateModifier(Enchantment aEnchantment, int aLevel) {
	            aEnchantment.func_151367_b(mPlayer, mEntity, aLevel);
	        }
	    }
	    
	    static final class BullshitIteratorB implements IBullshit {
	        public EntityLivingBase mPlayer;
	        public Entity mEntity;
	        BullshitIteratorB() {}
	        
	        @Override
			public void calculateModifier(Enchantment aEnchantment, int aLevel) {
	            aEnchantment.func_151368_a(mPlayer, mEntity, aLevel);
	        }
	    }
	    
	    interface IBullshit {
	        void calculateModifier(Enchantment aEnchantment, int aLevel);
	    }
	}
	
	public static class Reflection {
		public static Field getPublicField(Object aObject, String aField) {
			Field rField = null;
			try {
				rField = aObject.getClass().getDeclaredField(aField);
			} catch (Throwable e) {/*Do nothing*/}
			return rField;
		}
		
		public static Field getField(Object aObject, String aField) {
			Field rField = null;
			try {
				rField = aObject.getClass().getDeclaredField(aField);
				rField.setAccessible(T);
			} catch (Throwable e) {/*Do nothing*/}
			return rField;
		}
		
		public static Field getField(Class aObject, String aField) {
			Field rField = null;
			try {
				rField = aObject.getDeclaredField(aField);
				rField.setAccessible(T);
			} catch (Throwable e) {/*Do nothing*/}
			return rField;
		}
		
		public static Method getMethod(Class aObject, String aMethod, Class<?>... aParameterTypes) {
			Method rMethod = null;
			try {
				rMethod = aObject.getMethod(aMethod, aParameterTypes);
				rMethod.setAccessible(T);
			} catch (Throwable e) {/*Do nothing*/}
			return rMethod;
		}

		public static Method getMethod(Object aObject, String aMethod, Class<?>... aParameterTypes) {
			Method rMethod = null;
			try {
				rMethod = aObject.getClass().getMethod(aMethod, aParameterTypes);
				rMethod.setAccessible(T);
			} catch (Throwable e) {/*Do nothing*/}
			return rMethod;
		}

		public static Field getField(Object aObject, String aField, boolean aPrivate, boolean aLogErrors) {
			try {
				Field tField = (aObject instanceof Class)?((Class)aObject).getDeclaredField(aField):(aObject instanceof String)?Class.forName((String)aObject).getDeclaredField(aField):aObject.getClass().getDeclaredField(aField);
				if (aPrivate) tField.setAccessible(T);
				return tField;
			} catch (Throwable e) {
				if (aLogErrors) e.printStackTrace(GT_Log.err);
			}
			return null;
		}
		
		public static Object getFieldContent(Object aObject, String aField, boolean aPrivate, boolean aLogErrors) {
			try {
				Field tField = (aObject instanceof Class)?((Class)aObject).getDeclaredField(aField):(aObject instanceof String)?Class.forName((String)aObject).getDeclaredField(aField):aObject.getClass().getDeclaredField(aField);
				if (aPrivate) tField.setAccessible(T);
				return tField.get(aObject instanceof Class || aObject instanceof String ? null : aObject);
			} catch (Throwable e) {
				if (aLogErrors) e.printStackTrace(GT_Log.err);
			}
			return null;
		}
		
		public static Object callPublicMethod(Object aObject, String aMethod, Object... aParameters) {
			return callMethod(aObject, aMethod, F, F, T, aParameters);
		}

		public static Object callPrivateMethod(Object aObject, String aMethod, Object... aParameters) {
			return callMethod(aObject, aMethod, T, F, T, aParameters);
		}
		
		public static Object callMethod(Object aObject, String aMethod, boolean aPrivate, boolean aUseUpperCasedDataTypes, boolean aLogErrors, Object... aParameters) {
			try {
				Class<?>[] tParameterTypes = new Class<?>[aParameters.length];
				for (byte i = 0; i < aParameters.length; i++) {
					if (aParameters[i] instanceof Class) {
						tParameterTypes[i] = (Class)aParameters[i];
						aParameters[i] = null;
					} else {
						tParameterTypes[i] = aParameters[i].getClass();
					}
					if (!aUseUpperCasedDataTypes) {
						if (tParameterTypes[i] == Boolean.class	) tParameterTypes[i] = boolean.class;	else
						if (tParameterTypes[i] == Byte.class	) tParameterTypes[i] = byte.class;		else
						if (tParameterTypes[i] == Short.class	) tParameterTypes[i] = short.class;		else
						if (tParameterTypes[i] == Integer.class	) tParameterTypes[i] = int.class;		else
						if (tParameterTypes[i] == Long.class	) tParameterTypes[i] = long.class;		else
						if (tParameterTypes[i] == Float.class	) tParameterTypes[i] = float.class;		else
						if (tParameterTypes[i] == Double.class	) tParameterTypes[i] = double.class;
					}
				}
				
				Method tMethod = (aObject instanceof Class)?((Class)aObject).getMethod(aMethod, tParameterTypes):aObject.getClass().getMethod(aMethod, tParameterTypes);
				if (aPrivate) tMethod.setAccessible(T);
				return tMethod.invoke(aObject, aParameters);
			} catch (Throwable e) {
				if (aLogErrors) e.printStackTrace(GT_Log.err);
			}
			return null;
		}
		
		public static Object callConstructor(String aClass, int aConstructorIndex, Object aReplacementObject, boolean aLogErrors, Object... aParameters) {
			try {return callConstructor(Class.forName(aClass), aConstructorIndex, aReplacementObject, aLogErrors, aParameters);} catch (Throwable e) {if (aLogErrors) e.printStackTrace(GT_Log.err);} return aReplacementObject;
		}
		
		public static Object callConstructor(Class aClass, int aConstructorIndex, Object aReplacementObject, boolean aLogErrors, Object... aParameters) {
			if (aConstructorIndex < 0) {
				try {
					for (Constructor tConstructor : aClass.getConstructors()) {
						try {
							return tConstructor.newInstance(aParameters);
						} catch (Throwable e) {/*Do nothing*/}
					}
				} catch (Throwable e) {
					if (aLogErrors) e.printStackTrace(GT_Log.err);
				}
			} else {
				try {
					return aClass.getConstructors()[aConstructorIndex].newInstance(aParameters);
				} catch (Throwable e) {
					if (aLogErrors) e.printStackTrace(GT_Log.err);
				}
			}
			return aReplacementObject;
		}
		
	    public static String getClassName(Object aObject) {
	    	if (aObject == null) return "null";
	    	return aObject.getClass().getName().substring(aObject.getClass().getName().lastIndexOf(".")+1);
	    }
	}
	
	public static class Inventories {
		public static boolean TE_CHECK = F, BC_CHECK = F, CHECK_ALL = T;
		
		public static void checkAvailabilities() {
			if (CHECK_ALL) {
				try {
					Class tClass = cofh.api.transport.IItemConduit.class;
					tClass.getCanonicalName();
					TE_CHECK = T;
				} catch(Throwable e) {/**/}
				try {
					Class tClass = buildcraft.api.transport.IInjectable.class;
					tClass.getCanonicalName();
					BC_CHECK = T;
				} catch(Throwable e) {/**/}
				CHECK_ALL = F;
			}
		}
		
		public static boolean isConnectableNonInventoryPipe(Object aTileEntity, int aSide) {
			if (aTileEntity == null) return F;
			checkAvailabilities();
			if (TE_CHECK) if (aTileEntity instanceof cofh.api.transport.IItemConduit) return T;
			if (BC_CHECK) if (aTileEntity instanceof buildcraft.api.transport.IInjectable) return ((buildcraft.api.transport.IInjectable)aTileEntity).canInjectItems(FORGE_DIR[aSide]);
			return F;
		}
		
		/**
		 * Moves Stack from Inv-Slot to Inv-Slot, without checking if its even allowed.
		 * @return the Amount of moved Items
		 */
		public static byte moveStackIntoPipe(IInventory aTileEntity1, Object aTileEntity2, int[] aGrabSlots, byte aGrabFrom, byte aPutTo, List<ItemStack> aFilter, boolean aInvertFilter, byte aMaxTargetStackSize, byte aMinTargetStackSize, byte aMaxMoveAtOnce, byte aMinMoveAtOnce) {
			if (aTileEntity1 == null || aMaxTargetStackSize <= 0 || aMinTargetStackSize <= 0 || aMinTargetStackSize > aMaxTargetStackSize || aMaxMoveAtOnce <= 0 || aMinMoveAtOnce > aMaxMoveAtOnce) return 0;
			if (aTileEntity2 != null) {
				checkAvailabilities();
				if (TE_CHECK && aTileEntity2 instanceof cofh.api.transport.IItemConduit) {
					for (int i = 0; i < aGrabSlots.length; i++) {
						if (Stacks.inList(aFilter, aTileEntity1.getStackInSlot(aGrabSlots[i]), T, aInvertFilter)) {
							if (isAllowedToTakeFromSlot(aTileEntity1, aGrabSlots[i], aGrabFrom, aTileEntity1.getStackInSlot(aGrabSlots[i]))) {
								if (Math.max(aMinMoveAtOnce, aMinTargetStackSize) <= aTileEntity1.getStackInSlot(aGrabSlots[i]).stackSize) {
									ItemStack tStack = Stacks.copyAmount(Math.min(aTileEntity1.getStackInSlot(aGrabSlots[i]).stackSize, Math.min(aMaxMoveAtOnce, aMaxTargetStackSize)), aTileEntity1.getStackInSlot(aGrabSlots[i]));
									ItemStack rStack = ((cofh.api.transport.IItemConduit)aTileEntity2).insertItem(FORGE_DIR[aPutTo], Stacks.copy(tStack), F/*T*/);
									byte tMovedItemCount = (byte)(tStack.stackSize - (rStack == null ? 0 : rStack.stackSize));
									if (tMovedItemCount >= 1/*Math.max(aMinMoveAtOnce, aMinTargetStackSize)*/) {
										//((cofh.api.transport.IItemConduit)aTileEntity2).insertItem(FORGE_DIR[aPutTo], copyAmount(tMovedItemCount, tStack), F);
										aTileEntity1.decrStackSize(aGrabSlots[i], tMovedItemCount);
										aTileEntity1.markDirty();
										return tMovedItemCount;
									}
								}
							}
						}
					}
					return 0;
				}
				if (BC_CHECK && aTileEntity2 instanceof buildcraft.api.transport.IInjectable) {
					for (int i = 0; i < aGrabSlots.length; i++) {
						if (Stacks.inList(aFilter, aTileEntity1.getStackInSlot(aGrabSlots[i]), T, aInvertFilter)) {
							if (isAllowedToTakeFromSlot(aTileEntity1, aGrabSlots[i], aGrabFrom, aTileEntity1.getStackInSlot(aGrabSlots[i]))) {
								if (Math.max(aMinMoveAtOnce, aMinTargetStackSize) <= aTileEntity1.getStackInSlot(aGrabSlots[i]).stackSize) {
									ItemStack tStack = Stacks.copyAmount(Math.min(aTileEntity1.getStackInSlot(aGrabSlots[i]).stackSize, Math.min(aMaxMoveAtOnce, aMaxTargetStackSize)), aTileEntity1.getStackInSlot(aGrabSlots[i]));
									byte tMovedItemCount = (byte)((buildcraft.api.transport.IInjectable)aTileEntity2).injectItem(Stacks.copy(tStack), F, FORGE_DIR[aPutTo], null);
									if (tMovedItemCount >= Math.max(aMinMoveAtOnce, aMinTargetStackSize)) {
										tMovedItemCount = (byte)(((buildcraft.api.transport.IInjectable)aTileEntity2).injectItem(Stacks.copyAmount(tMovedItemCount, tStack), T, FORGE_DIR[aPutTo], null));
										aTileEntity1.decrStackSize(aGrabSlots[i], tMovedItemCount);
										aTileEntity1.markDirty();
										return tMovedItemCount;
									}
								}
							}
						}
					}
					return 0;
				}
			}
			
			if (aTileEntity1 instanceof TileEntity && SIDES_VALID[aGrabFrom] && OPPOSITES[aGrabFrom] == aPutTo) {
				int tX = ((TileEntity)aTileEntity1).xCoord + OFFSETS_X[aGrabFrom], tY = ((TileEntity)aTileEntity1).yCoord + OFFSETS_Y[aGrabFrom], tZ = ((TileEntity)aTileEntity1).zCoord + OFFSETS_Z[aGrabFrom];
				if (!Worlds.hasBlockHitBox(((TileEntity)aTileEntity1).getWorldObj(), tX, tY, tZ)) {
					for (int i = 0; i < aGrabSlots.length; i++) {
						if (Stacks.inList(aFilter, aTileEntity1.getStackInSlot(aGrabSlots[i]), T, aInvertFilter)) {
							if (isAllowedToTakeFromSlot(aTileEntity1, aGrabSlots[i], aGrabFrom, aTileEntity1.getStackInSlot(aGrabSlots[i]))) {
								if (Math.max(aMinMoveAtOnce, aMinTargetStackSize) <= aTileEntity1.getStackInSlot(aGrabSlots[i]).stackSize) {
									ItemStack tStack = Stacks.copyAmount(Math.min(aTileEntity1.getStackInSlot(aGrabSlots[i]).stackSize, Math.min(aMaxMoveAtOnce, aMaxTargetStackSize)), aTileEntity1.getStackInSlot(aGrabSlots[i]));
									EntityItem tEntity = new EntityItem(((TileEntity)aTileEntity1).getWorldObj(), tX+0.5, tY+0.5, tZ+0.5, tStack);
									tEntity.motionX = tEntity.motionY = tEntity.motionZ = 0;
									((TileEntity)aTileEntity1).getWorldObj().spawnEntityInWorld(tEntity);
									aTileEntity1.decrStackSize(aGrabSlots[i], tStack.stackSize);
									aTileEntity1.markDirty();
									return (byte)tStack.stackSize;
								}
							}
						}
					}
				}
			}
			return 0;
		}
		
		/**
		 * Moves Stack from Inv-Slot to Inv-Slot, without checking if its even allowed. (useful for internal Inventory Operations)
		 * @return the Amount of moved Items
		 */
		public static byte moveStackFromSlotAToSlotB(IInventory aTileEntity1, IInventory aTileEntity2, int aGrabFrom, int aPutTo, byte aMaxTargetStackSize, byte aMinTargetStackSize, byte aMaxMoveAtOnce, byte aMinMoveAtOnce) {
			if (aTileEntity1 == null || aTileEntity2 == null || aMaxTargetStackSize <= 0 || aMinTargetStackSize <= 0 || aMinTargetStackSize > aMaxTargetStackSize || aMaxMoveAtOnce <= 0 || aMinMoveAtOnce > aMaxMoveAtOnce) return 0;
			
			ItemStack tStack1 = aTileEntity1.getStackInSlot(aGrabFrom), tStack2 = aTileEntity2.getStackInSlot(aPutTo), tStack3 = null;
			if (tStack1 != null) {
				if (tStack2 != null && !Stacks.equal(tStack1, tStack2)) return 0;
				tStack3 = Stacks.copy(tStack1);
				aMaxTargetStackSize = (byte)Math.min(aMaxTargetStackSize, Math.min(tStack3.getMaxStackSize(), Math.min(tStack2==null?Integer.MAX_VALUE:tStack2.getMaxStackSize(), aTileEntity2.getInventoryStackLimit())));
				tStack3.stackSize = Math.min(tStack3.stackSize, aMaxTargetStackSize - (tStack2 == null?0:tStack2.stackSize));
				if (tStack3.stackSize > aMaxMoveAtOnce) tStack3.stackSize = aMaxMoveAtOnce;
				if (tStack3.stackSize + (tStack2==null?0:tStack2.stackSize) >= Math.min(tStack3.getMaxStackSize(), aMinTargetStackSize) && tStack3.stackSize >= aMinMoveAtOnce) {
					tStack3 = aTileEntity1.decrStackSize(aGrabFrom, tStack3.stackSize);
					aTileEntity1.markDirty();
					if (tStack3 != null) {
						if (tStack2 == null) {
							aTileEntity2.setInventorySlotContents(aPutTo, Stacks.copy(tStack3));
							aTileEntity2.markDirty();
						} else {
							tStack2.stackSize += tStack3.stackSize;
							aTileEntity2.markDirty();
						}
						return (byte)tStack3.stackSize;
					}
				}
			}
			return 0;
		}
		
		public static boolean isAllowedToTakeFromSlot(IInventory aTileEntity, int aSlot, byte aSide, ItemStack aStack) {
			if (SIDES_INVALID[aSide]) {
				return isAllowedToTakeFromSlot(aTileEntity, aSlot, (byte)0, aStack)
					|| isAllowedToTakeFromSlot(aTileEntity, aSlot, (byte)1, aStack)
					|| isAllowedToTakeFromSlot(aTileEntity, aSlot, (byte)2, aStack)
					|| isAllowedToTakeFromSlot(aTileEntity, aSlot, (byte)3, aStack)
					|| isAllowedToTakeFromSlot(aTileEntity, aSlot, (byte)4, aStack)
					|| isAllowedToTakeFromSlot(aTileEntity, aSlot, (byte)5, aStack);
			}
			if (aTileEntity instanceof ISidedInventory) return ((ISidedInventory)aTileEntity).canExtractItem(aSlot, aStack, aSide);
			return T;
		}
		
		public static boolean isAllowedToPutIntoSlot(IInventory aTileEntity, int aSlot, byte aSide, ItemStack aStack, byte aMaxStackSize) {
			ItemStack tStack = aTileEntity.getStackInSlot(aSlot);
			if (tStack != null && (!Stacks.equal(tStack, aStack) || tStack.stackSize >= tStack.getMaxStackSize())) return F;
			if (SIDES_INVALID[aSide]) {
				return isAllowedToPutIntoSlot(aTileEntity, aSlot, (byte)0, aStack, aMaxStackSize)
					|| isAllowedToPutIntoSlot(aTileEntity, aSlot, (byte)1, aStack, aMaxStackSize)
					|| isAllowedToPutIntoSlot(aTileEntity, aSlot, (byte)2, aStack, aMaxStackSize)
					|| isAllowedToPutIntoSlot(aTileEntity, aSlot, (byte)3, aStack, aMaxStackSize)
					|| isAllowedToPutIntoSlot(aTileEntity, aSlot, (byte)4, aStack, aMaxStackSize)
					|| isAllowedToPutIntoSlot(aTileEntity, aSlot, (byte)5, aStack, aMaxStackSize);
			}
			if (aTileEntity instanceof ISidedInventory && !((ISidedInventory)aTileEntity).canInsertItem(aSlot, aStack, aSide)) return F;
			return aTileEntity.isItemValidForSlot(aSlot, aStack);
		}
		
		/**
		 * Moves Stack from Inv-Side to Inv-Side.
		 * @return the Amount of moved Items
		 */
		public static byte moveOneItemStack(Object aTileEntity1, Object aTileEntity2, byte aGrabFrom, byte aPutTo, List<ItemStack> aFilter, boolean aInvertFilter, byte aMaxTargetStackSize, byte aMinTargetStackSize, byte aMaxMoveAtOnce, byte aMinMoveAtOnce) {
			if (aTileEntity1 != null && aTileEntity1 instanceof IInventory) return moveOneItemStack((IInventory)aTileEntity1, aTileEntity2, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, T);
			return 0;
		}
		
		/**
		 * This is only because I needed an additional Parameter for the Double Chest Check.
		 */
		private static byte moveOneItemStack(IInventory aTileEntity1, Object aTileEntity2, byte aGrabFrom, byte aPutTo, List<ItemStack> aFilter, boolean aInvertFilter, byte aMaxTargetStackSize, byte aMinTargetStackSize, byte aMaxMoveAtOnce, byte aMinMoveAtOnce, boolean aDoCheckChests) {
			if (aTileEntity1 == null || aMaxTargetStackSize <= 0 || aMinTargetStackSize <= 0 || aMaxMoveAtOnce <= 0 || aMinTargetStackSize > aMaxTargetStackSize || aMinMoveAtOnce > aMaxMoveAtOnce) return 0;
			
			int[] tGrabSlots = null;
			if (aTileEntity1 instanceof ISidedInventory) tGrabSlots = ((ISidedInventory)aTileEntity1).getAccessibleSlotsFromSide(aGrabFrom);
			if (tGrabSlots == null) {
				tGrabSlots = new int[aTileEntity1.getSizeInventory()];
				for (int i = 0; i < tGrabSlots.length; i++) tGrabSlots[i] = i;
			}
			
			if (aTileEntity2 != null && aTileEntity2 instanceof IInventory) {
				int[] tPutSlots = null;
				if (aTileEntity2 instanceof ISidedInventory) tPutSlots = ((ISidedInventory)aTileEntity2).getAccessibleSlotsFromSide(aPutTo);
				
				if (tPutSlots == null) {
					tPutSlots = new int[((IInventory)aTileEntity2).getSizeInventory()];
					for (int i = 0; i < tPutSlots.length; i++) tPutSlots[i] = i;
				}
				
				for (int i = 0; i < tGrabSlots.length; i++) {
					for (int j = 0; j < tPutSlots.length; j++) {
						if (Stacks.inList(aFilter, aTileEntity1.getStackInSlot(tGrabSlots[i]), T, aInvertFilter)) {
							if (isAllowedToTakeFromSlot(aTileEntity1, tGrabSlots[i], aGrabFrom, aTileEntity1.getStackInSlot(tGrabSlots[i]))) {
								if (isAllowedToPutIntoSlot((IInventory)aTileEntity2, tPutSlots[j], aPutTo, aTileEntity1.getStackInSlot(tGrabSlots[i]), aMaxTargetStackSize)) {
									byte tMovedItemCount = moveStackFromSlotAToSlotB(aTileEntity1, (IInventory)aTileEntity2, tGrabSlots[i], tPutSlots[j], aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce);
									if (tMovedItemCount > 0) return tMovedItemCount;
								}
							}
						}
					}
				}
				
				if (aDoCheckChests && aTileEntity1 instanceof TileEntityChest) {
					TileEntityChest tTileEntity1 = (TileEntityChest)aTileEntity1;
					if (tTileEntity1.adjacentChestChecked) {
						byte tAmount = 0;
						if (tTileEntity1.adjacentChestXNeg != null) {
							tAmount = moveOneItemStack(tTileEntity1.adjacentChestXNeg, aTileEntity2, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, F);
						} else if (tTileEntity1.adjacentChestZNeg != null) {
							tAmount = moveOneItemStack(tTileEntity1.adjacentChestZNeg, aTileEntity2, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, F);
						} else if (tTileEntity1.adjacentChestXPos != null) {
							tAmount = moveOneItemStack(tTileEntity1.adjacentChestXPos, aTileEntity2, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, F);
						} else if (tTileEntity1.adjacentChestZPos != null) {
							tAmount = moveOneItemStack(tTileEntity1.adjacentChestZPos, aTileEntity2, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, F);
						}
						if (tAmount != 0) return tAmount;
					}
				}
				if (aDoCheckChests && aTileEntity2 instanceof TileEntityChest) {
					TileEntityChest tTileEntity2 = (TileEntityChest)aTileEntity2;
					if (tTileEntity2.adjacentChestChecked) {
						byte tAmount = 0;
						if (tTileEntity2.adjacentChestXNeg != null) {
							tAmount = moveOneItemStack(aTileEntity1, tTileEntity2.adjacentChestXNeg, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, F);
						} else if (tTileEntity2.adjacentChestZNeg != null) {
							tAmount = moveOneItemStack(aTileEntity1, tTileEntity2.adjacentChestZNeg, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, F);
						} else if (tTileEntity2.adjacentChestXPos != null) {
							tAmount = moveOneItemStack(aTileEntity1, tTileEntity2.adjacentChestXPos, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, F);
						} else if (tTileEntity2.adjacentChestZPos != null) {
							tAmount = moveOneItemStack(aTileEntity1, tTileEntity2.adjacentChestZPos, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce, F);
						}
						if (tAmount != 0) return tAmount;
					}
				}
			}
			
			return moveStackIntoPipe(aTileEntity1, aTileEntity2, tGrabSlots, aGrabFrom, aPutTo, aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce);
		}
		
		/**
		 * Moves Stack from Inv-Side to Inv-Slot.
		 * @return the Amount of moved Items
		 */
		public static byte moveOneItemStackIntoSlot(Object aTileEntity1, Object aTileEntity2, byte aGrabFrom, int aPutTo, List<ItemStack> aFilter, boolean aInvertFilter, byte aMaxTargetStackSize, byte aMinTargetStackSize, byte aMaxMoveAtOnce, byte aMinMoveAtOnce) {
			if (aTileEntity1 == null || !(aTileEntity1 instanceof IInventory) || aPutTo < 0 || aMaxTargetStackSize <= 0 || aMinTargetStackSize <= 0 || aMaxMoveAtOnce <= 0 || aMinTargetStackSize > aMaxTargetStackSize || aMinMoveAtOnce > aMaxMoveAtOnce) return 0;
			
			int[] tGrabSlots = null;
			if (aTileEntity1 instanceof ISidedInventory) tGrabSlots = ((ISidedInventory)aTileEntity1).getAccessibleSlotsFromSide(aGrabFrom);
			if (tGrabSlots == null) {
				tGrabSlots = new int[((IInventory)aTileEntity1).getSizeInventory()];
				for (int i = 0; i < tGrabSlots.length; i++) tGrabSlots[i] = i;
			}
			
			if (aTileEntity2 != null && aTileEntity2 instanceof IInventory) {
				for (int i = 0; i < tGrabSlots.length; i++) {
					if (Stacks.inList(aFilter, ((IInventory)aTileEntity1).getStackInSlot(tGrabSlots[i]), T, aInvertFilter)) {
						if (isAllowedToTakeFromSlot((IInventory)aTileEntity1, tGrabSlots[i], aGrabFrom, ((IInventory)aTileEntity1).getStackInSlot(tGrabSlots[i]))) {
							if (isAllowedToPutIntoSlot((IInventory)aTileEntity2, aPutTo, SIDE_ANY, ((IInventory)aTileEntity1).getStackInSlot(tGrabSlots[i]), aMaxTargetStackSize)) {
								byte tMovedItemCount = moveStackFromSlotAToSlotB((IInventory)aTileEntity1, (IInventory)aTileEntity2, tGrabSlots[i], aPutTo, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce);
								if (tMovedItemCount > 0) return tMovedItemCount;
							}
						}
					}
				}
			}
			
			moveStackIntoPipe(((IInventory)aTileEntity1), aTileEntity2, tGrabSlots, aGrabFrom, OPPOSITES[aGrabFrom], aFilter, aInvertFilter, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce);
			return 0;
		}
		
		/**
		 * Moves Stack from Inv-Slot to Inv-Slot.
		 * @return the Amount of moved Items
		 */
		public static byte moveFromSlotToSlot(IInventory aTileEntity1, IInventory aTileEntity2, int aGrabFrom, int aPutTo, List<ItemStack> aFilter, boolean aInvertFilter, byte aMaxTargetStackSize, byte aMinTargetStackSize, byte aMaxMoveAtOnce, byte aMinMoveAtOnce) {
			if (aTileEntity1 == null || aTileEntity2 == null || aGrabFrom < 0 || aPutTo < 0 || aMaxTargetStackSize <= 0 || aMinTargetStackSize <= 0 || aMaxMoveAtOnce <= 0 || aMinTargetStackSize > aMaxTargetStackSize || aMinMoveAtOnce > aMaxMoveAtOnce) return 0;
			if (Stacks.inList(aFilter, aTileEntity1.getStackInSlot(aGrabFrom), T, aInvertFilter)) {
				if (isAllowedToTakeFromSlot(aTileEntity1, aGrabFrom, (byte)6, aTileEntity1.getStackInSlot(aGrabFrom))) {
					if (isAllowedToPutIntoSlot(aTileEntity2, aPutTo, (byte)6, aTileEntity1.getStackInSlot(aGrabFrom), aMaxTargetStackSize)) {
						byte tMovedItemCount = moveStackFromSlotAToSlotB(aTileEntity1, aTileEntity2, aGrabFrom, aPutTo, aMaxTargetStackSize, aMinTargetStackSize, aMaxMoveAtOnce, aMinMoveAtOnce);
						if (tMovedItemCount > 0) return tMovedItemCount;
					}
				}
			}
			return 0;
		}
		
		public static void removeNullStacksFromInventory(IInventory aInventory) {
			if (aInventory != null) for (int i = 0, j = aInventory.getSizeInventory(); i < j; i++) {
				ItemStack tStack = aInventory.getStackInSlot(i);
				if (tStack != null && (tStack.stackSize == 0 || tStack.getItem() == null)) aInventory.setInventorySlotContents(i, null);
			}
		}
		
		public static boolean addStackToPlayerInventory(EntityPlayer aPlayer, ItemStack aStack) {
			if (aPlayer != null) {
				for (int i = 0; i < 36; i++) if (i != aPlayer.inventory.currentItem) {
					ItemStack tStack = aPlayer.inventory.getStackInSlot(i);
					if (Stacks.equal(tStack, aStack) && aStack.stackSize + tStack.stackSize <= tStack.getMaxStackSize()) {
						tStack.stackSize += aStack.stackSize;
						if (aPlayer.openContainer != null) aPlayer.openContainer.detectAndSendChanges();
						return true;
					}
				}
				for (int i = 0; i < 36; i++) if (i != aPlayer.inventory.currentItem) {
					ItemStack tStack = aPlayer.inventory.getStackInSlot(i);
					if (tStack == null) {
						aPlayer.inventory.setInventorySlotContents(i, aStack);
						if (aPlayer.openContainer != null) aPlayer.openContainer.detectAndSendChanges();
						return true;
					}
				}
				ItemStack tStack = aPlayer.inventory.getStackInSlot(aPlayer.inventory.currentItem);
				if (tStack == null) {
					aPlayer.inventory.setInventorySlotContents(aPlayer.inventory.currentItem, aStack);
					if (aPlayer.openContainer != null) aPlayer.openContainer.detectAndSendChanges();
					return true;
				} else if (Stacks.equal(tStack, aStack) && aStack.stackSize + tStack.stackSize <= tStack.getMaxStackSize()) {
					tStack.stackSize += aStack.stackSize;
					if (aPlayer.openContainer != null) aPlayer.openContainer.detectAndSendChanges();
					return true;
				}
			}
			return false;
		}
		
		public static ItemStack getProjectile(TagData aProjectileType, IInventory aInventory) {
			if (aInventory != null) for (int i = 0, j = aInventory.getSizeInventory(); i < j; i++) {
				ItemStack rStack = aInventory.getStackInSlot(i);
				if (Stacks.valid(rStack) && rStack.getItem() instanceof IItemProjectile && ((IItemProjectile)rStack.getItem()).hasProjectile(aProjectileType, rStack)) return Stacks.update(rStack);
			}
			return null;
		}
	}
	
	public static class Sounds {
		public static List<PlayedSound> sPlayedSoundMap = new ArrayListNoNulls<PlayedSound>();
		
		public static boolean play(String aSoundName, int aTimeUntilNextSound, float aSoundStrength) {
			return play(aSoundName, aTimeUntilNextSound, aSoundStrength, GT_API.api_proxy.getThePlayer());
		}
		
		public static boolean play(String aSoundName, int aTimeUntilNextSound, float aSoundStrength, Entity aEntity) {
			if (aEntity == null) return F;
			return play(aSoundName, aTimeUntilNextSound, aSoundStrength, MathHelper.floor_double(aEntity.posX), MathHelper.floor_double(aEntity.posY), MathHelper.floor_double(aEntity.posZ));
		}
		
		public static boolean play(String aSoundName, int aTimeUntilNextSound, float aSoundStrength, int aX, int aY, int aZ) {
			return play(aSoundName, aTimeUntilNextSound, aSoundStrength, 0.9F + RANDOM.nextFloat() * 0.2F, aX, aY, aZ);
		}
		
		public static boolean play(String aSoundName, int aTimeUntilNextSound, float aSoundStrength, float aSoundModulation, int aX, int aY, int aZ) {
			if (Code.stringInvalid(aSoundName) || !FMLCommonHandler.instance().getEffectiveSide().isClient() || GT_API.api_proxy.getThePlayer() == null || !GT_API.api_proxy.getThePlayer().worldObj.isRemote) return F;
			if (GregTech_API.sMultiThreadedSounds)
				new Thread(new GT_Runnable_Sound(GT_API.api_proxy.getThePlayer().worldObj, MathHelper.floor_double(aX), MathHelper.floor_double(aY), MathHelper.floor_double(aZ), aTimeUntilNextSound, aSoundName, aSoundStrength, aSoundModulation), "Sound Effect").start();
			else
				new GT_Runnable_Sound(GT_API.api_proxy.getThePlayer().worldObj, MathHelper.floor_double(aX), MathHelper.floor_double(aY), MathHelper.floor_double(aZ), aTimeUntilNextSound, aSoundName, aSoundStrength, aSoundModulation).run();
			return T;
		}
		
		public static boolean send(World aWorld, String aSoundName, float aSoundStrength, float aSoundModulation, int aX, int aY, int aZ) {
			if (Code.stringInvalid(aSoundName) || aWorld == null || aWorld.isRemote) return F;
	        NW_API.sendToAllPlayersInRange(new PacketSound(aSoundName, aSoundStrength, aSoundModulation, aX, aY, aZ), aWorld, aX, aZ);
			return T;
		}
	}
	
	public static class Entities {
		/** Sends a Chat Message to a Player */
		public static void chat(Object aPlayer, String aChatMessage) {
			if (aPlayer instanceof EntityPlayerMP && aChatMessage != null) {
				((EntityPlayerMP)aPlayer).addChatComponentMessage(new ChatComponentText(aChatMessage));
			}
		}
		
	    public static boolean isWearingFullFrostHazmat(EntityLivingBase aEntity) {
			for (byte i = 1; i < 5; i++) if (!Stacks.inList(aEntity.getEquipmentInSlot(i), GregTech_API.sFrostHazmatList)) return F;
			return T;
	    }
	    
	    public static boolean isWearingFullHeatHazmat(EntityLivingBase aEntity) {
			for (byte i = 1; i < 5; i++) if (!Stacks.inList(aEntity.getEquipmentInSlot(i), GregTech_API.sHeatHazmatList)) return F;
			return T;
	    }
	    
	    public static boolean isWearingFullBioHazmat(EntityLivingBase aEntity) {
			for (byte i = 1; i < 5; i++) if (!Stacks.inList(aEntity.getEquipmentInSlot(i), GregTech_API.sBioHazmatList)) return F;
			return T;
	    }
	    
	    public static boolean isWearingFullRadioHazmat(EntityLivingBase aEntity) {
			for (byte i = 1; i < 5; i++) if (!Stacks.inList(aEntity.getEquipmentInSlot(i), GregTech_API.sRadioHazmatList)) return F;
			return T;
	    }
	    
	    public static boolean isWearingFullElectroHazmat(EntityLivingBase aEntity) {
			for (byte i = 1; i < 5; i++) if (!Stacks.inList(aEntity.getEquipmentInSlot(i), GregTech_API.sElectroHazmatList)) return F;
			return T;
	    }
	    
	    public static boolean isWearingFullGasHazmat(EntityLivingBase aEntity) {
			for (byte i = 1; i < 5; i++) if (!Stacks.inList(aEntity.getEquipmentInSlot(i), GregTech_API.sGasHazmatList)) return F;
			return T;
	    }
	    
	    public static float getHeatDamageFromItem(ItemStack aStack) {
	    	OreDictItemData tData = OM.data(aStack, T);
	    	return tData==null?0:(tData.mPrefix==null?0:tData.mPrefix.mHeatDamage) + (tData.hasValidMaterialData()?tData.mMaterial.mMaterial.mHeatDamage:0);
	    }
	    
	    public static int getRadioactivityLevel(ItemStack aStack) {
	    	OreDictItemData tData = OM.data(aStack, T);
	    	if (tData != null && tData.hasValidMaterialData()) {
	    		if (tData.mMaterial.mMaterial.mEnchantmentArmors instanceof Enchantment_Radioactivity) return tData.mMaterial.mMaterial.mEnchantmentArmorsLevel;
	        	if (tData.mMaterial.mMaterial.mEnchantmentTools instanceof Enchantment_Radioactivity) return tData.mMaterial.mMaterial.mEnchantmentToolsLevel;
	    	}
	    	return EnchantmentHelper.getEnchantmentLevel(Enchantment_Radioactivity.INSTANCE.effectId, aStack);
	    }
	    
	    public static boolean isImmuneToBreathingGasses(EntityLivingBase aEntity) {
	    	return isWearingFullGasHazmat(aEntity);
	    }
	    
	    public static boolean applyHeatDamage(EntityLivingBase aEntity, float aDamage) {
	    	if (aDamage > 0 && aEntity != null && aEntity.getActivePotionEffect(Potion.fireResistance) == null && !isWearingFullHeatHazmat(aEntity)) {
	        	aEntity.attackEntityFrom(DamageSources.getHeatDamage(), aDamage);
	    		return T;
	    	}
	    	return F;
	    }
	    
	    public static boolean applyFrostDamage(EntityLivingBase aEntity, float aDamage) {
	    	if (aDamage > 0 && aEntity != null && !isWearingFullFrostHazmat(aEntity)) {
	        	aEntity.attackEntityFrom(DamageSources.getFrostDamage(), aDamage);
	    		return T;
	    	}
	    	return F;
	    }
	    
	    public static boolean applyElectricityDamage(EntityLivingBase aEntity, long aVoltage, long aAmperage) {
	    	long aDamage = Code.tier(aVoltage) * aAmperage * 4;
	    	if (aDamage > 0 && aEntity != null && !isWearingFullElectroHazmat(aEntity)) {
	        	aEntity.attackEntityFrom(DamageSources.getElectricDamage(), aDamage);
	    		return T;
	    	}
	    	return F;
	    }
	    
	    public static boolean applyRadioactivity(EntityLivingBase aEntity, int aLevel, int aAmountOfItems) {
	    	if (aLevel > 0 && aEntity != null && aEntity.getCreatureAttribute() != EnumCreatureAttribute.UNDEAD && aEntity.getCreatureAttribute() != EnumCreatureAttribute.ARTHROPOD && !isWearingFullRadioHazmat(aEntity)) {
	        	PotionEffect tEffect = null;
	    		aEntity.addPotionEffect(new PotionEffect(Potion.moveSlowdown.id	, aLevel * 140 * aAmountOfItems + Math.max(0, ((tEffect = aEntity.getActivePotionEffect(Potion.moveSlowdown		))==null?0:tEffect.getDuration())), Math.max(0, (5*aLevel) / 7)));
	        	aEntity.addPotionEffect(new PotionEffect(Potion.digSlowdown.id	, aLevel * 150 * aAmountOfItems + Math.max(0, ((tEffect = aEntity.getActivePotionEffect(Potion.digSlowdown		))==null?0:tEffect.getDuration())), Math.max(0, (5*aLevel) / 7)));
	        	aEntity.addPotionEffect(new PotionEffect(Potion.confusion.id	, aLevel * 130 * aAmountOfItems + Math.max(0, ((tEffect = aEntity.getActivePotionEffect(Potion.confusion		))==null?0:tEffect.getDuration())), Math.max(0, (5*aLevel) / 7)));
	        	aEntity.addPotionEffect(new PotionEffect(Potion.weakness.id		, aLevel * 150 * aAmountOfItems + Math.max(0, ((tEffect = aEntity.getActivePotionEffect(Potion.weakness			))==null?0:tEffect.getDuration())), Math.max(0, (5*aLevel) / 7)));
	        	aEntity.addPotionEffect(new PotionEffect(Potion.hunger.id		, aLevel * 130 * aAmountOfItems + Math.max(0, ((tEffect = aEntity.getActivePotionEffect(Potion.hunger			))==null?0:tEffect.getDuration())), Math.max(0, (5*aLevel) / 7)));
	        	aEntity.addPotionEffect(new PotionEffect(24 /* IC2 Radiation */	, aLevel * 180 * aAmountOfItems + Math.max(0, ((tEffect = aEntity.getActivePotionEffect(Potion.potionTypes[24]	))==null?0:tEffect.getDuration())), Math.max(0, (5*aLevel) / 7)));
	        	return T;
	    	}
	    	return F;
	    }
	    
	    public static boolean getPotion(EntityLivingBase aPlayer, int aPotionIndex) {
	        try  {
	        	Field tPotionHashmap = null;
	        	
	            Field[] var3 = EntityLiving.class.getDeclaredFields();
	            int var4 = var3.length;
	            
	            for (int var5 = 0; var5 < var4; ++var5) {
	                Field var6 = var3[var5];
	                if (var6.getType() == HashMap.class) {
	                    tPotionHashmap = var6;
	                    tPotionHashmap.setAccessible(T);
	                    break;
	                }
	            }
	            
	            if (tPotionHashmap != null) return ((HashMap)tPotionHashmap.get(aPlayer)).get(Integer.valueOf(aPotionIndex)) != null;
	        } catch (Throwable e) {
	        	if (D1) e.printStackTrace(GT_Log.err);
	        }
	    	return F;
	    }
		
	    public static void removePotion(EntityLivingBase aPlayer, int aPotionIndex) {
	        try  {
	        	Field tPotionHashmap = null;
	        	
	            Field[] var3 = EntityLiving.class.getDeclaredFields();
	            int var4 = var3.length;

	            for (int var5 = 0; var5 < var4; ++var5) {
	                Field var6 = var3[var5];
	                if (var6.getType() == HashMap.class) {
	                    tPotionHashmap = var6;
	                    tPotionHashmap.setAccessible(T);
	                    break;
	                }
	            }

	            if (tPotionHashmap != null) ((HashMap)tPotionHashmap.get(aPlayer)).remove(Integer.valueOf(aPotionIndex));
	        } catch (Throwable e) {
	        	if (D1) e.printStackTrace(GT_Log.err);
	        }
	    }
		
		public static boolean getFullInvisibility(EntityPlayer aPlayer) {
			try {
				if (aPlayer.isInvisible()) {
					for (int i = 0; i < 4; i++) {
						if (aPlayer.inventory.armorInventory[i] != null) {
							if (aPlayer.inventory.armorInventory[i].getItem() instanceof GT_EnergyArmor_Item) {
								if ((((GT_EnergyArmor_Item)aPlayer.inventory.armorInventory[i].getItem()).mSpecials & 512) != 0) {
									if (GT_ModHandler.canUseElectricItem(aPlayer.inventory.armorInventory[i], 10000)) {
										return T;
									}
								}
							}
						}
					}
				}
			} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
			return F;
		}
	}
	
	public static class Worlds {
		public static boolean isOpaqueBlock(World aWorld, int aX, int aY, int aZ) {
			return aWorld.getBlock(aX, aY, aZ).isOpaqueCube();
		}
		
		public static boolean isAirBlock(World aWorld, int aX, int aY, int aZ) {
			return aWorld.getBlock(aX, aY, aZ).isAir(aWorld, aX, aY, aZ);
		}
		
		public static ItemStack suckOneItemStackAt(World aWorld, double aX, double aY, double aZ, double aL, double aH, double aW) {
			for (EntityItem tItem : (ArrayList<EntityItem>)aWorld.getEntitiesWithinAABB(EntityItem.class, AxisAlignedBB.getBoundingBox(aX, aY, aZ, aX+aL, aY+aH, aZ+aW))) {
				if (!tItem.isDead) {
					aWorld.removeEntity(tItem);
					tItem.setDead();
					return tItem.getEntityItem();
				}
			}
			return null;
		}
		
		public static boolean isSideObstructed(World aWorld, int aX, int aY, int aZ, byte aSide) {
			aX += OFFSETS_X[aSide]; aY += OFFSETS_Y[aSide]; aZ += OFFSETS_Z[aSide];
			Block tBlock = aWorld.getBlock(aX, aY, aZ);
    		AxisAlignedBB tBoundingBox = tBlock.getCollisionBoundingBoxFromPool(aWorld, aX, aY, aZ);
    		if (tBoundingBox == null) return false;
    		switch(aSide) {
    		case 0: return tBoundingBox.maxY-aY > PIXELS_NEG[4] && tBoundingBox.maxX-aX > PIXELS_POS[2] && tBoundingBox.minX-aX < PIXELS_NEG[2] && tBoundingBox.maxZ-aZ > PIXELS_POS[2] && tBoundingBox.minZ-aZ < PIXELS_NEG[2];
    		case 1: return tBoundingBox.minY-aY < PIXELS_POS[4] && tBoundingBox.maxX-aX > PIXELS_POS[2] && tBoundingBox.minX-aX < PIXELS_NEG[2] && tBoundingBox.maxZ-aZ > PIXELS_POS[2] && tBoundingBox.minZ-aZ < PIXELS_NEG[2];
    		case 2: return tBoundingBox.maxZ-aZ > PIXELS_NEG[4] && tBoundingBox.maxX-aX > PIXELS_POS[2] && tBoundingBox.minX-aX < PIXELS_NEG[2] && tBoundingBox.maxY-aY > PIXELS_POS[2] && tBoundingBox.minY-aY < PIXELS_NEG[2];
    		case 3: return tBoundingBox.minZ-aZ < PIXELS_POS[4] && tBoundingBox.maxX-aX > PIXELS_POS[2] && tBoundingBox.minX-aX < PIXELS_NEG[2] && tBoundingBox.maxY-aY > PIXELS_POS[2] && tBoundingBox.minY-aY < PIXELS_NEG[2];
    		case 4: return tBoundingBox.maxX-aX > PIXELS_NEG[4] && tBoundingBox.maxZ-aZ > PIXELS_POS[2] && tBoundingBox.minZ-aZ < PIXELS_NEG[2] && tBoundingBox.maxY-aY > PIXELS_POS[2] && tBoundingBox.minY-aY < PIXELS_NEG[2];
    		case 5: return tBoundingBox.minX-aX < PIXELS_POS[4] && tBoundingBox.maxZ-aZ > PIXELS_POS[2] && tBoundingBox.minZ-aZ < PIXELS_NEG[2] && tBoundingBox.maxY-aY > PIXELS_POS[2] && tBoundingBox.minY-aY < PIXELS_NEG[2];
    		}
    		return false;
		}
		
		/**
		 * This checks if the Dimension is really a Dimension and not another Planet or something.
		 * Used for my Teleporter.
		 */
		public static boolean isRealDimension(int aDimensionID) {
			try {
				if (DimensionManager.getProvider(aDimensionID).getClass().getName().contains("com.xcompwiz.mystcraft")) return T;
			} catch (Throwable e) {/*Do nothing*/}
			try {
				if (DimensionManager.getProvider(aDimensionID).getClass().getName().contains("TwilightForest")) return T;
			} catch (Throwable e) {/*Do nothing*/}
			return GregTech_API.sDimensionalList.contains(aDimensionID);
		}
		
		public static boolean moveEntityToDimensionAtCoords(Entity aEntity, int aDimension, double aX, double aY, double aZ) {
			WorldServer tTargetWorld = DimensionManager.getWorld(aDimension), tOriginalWorld = DimensionManager.getWorld(aEntity.worldObj.provider.dimensionId);
			if (tTargetWorld != null && tOriginalWorld != null && tTargetWorld != tOriginalWorld) {
				if (aEntity.ridingEntity != null) aEntity.mountEntity(null);
				if (aEntity.riddenByEntity != null) aEntity.riddenByEntity.mountEntity(null);
				
				if (aEntity instanceof EntityPlayerMP) {
					EntityPlayerMP aPlayer = (EntityPlayerMP)aEntity;
			        aPlayer.dimension = aDimension;
			        aPlayer.playerNetServerHandler.sendPacket(new S07PacketRespawn(aPlayer.dimension, aPlayer.worldObj.difficultySetting, aPlayer.worldObj.getWorldInfo().getTerrainType(), aPlayer.theItemInWorldManager.getGameType()));
			        tOriginalWorld.removePlayerEntityDangerously(aPlayer);
			        aPlayer.isDead = F;
			        aPlayer.setWorld(tTargetWorld);
			        MinecraftServer.getServer().getConfigurationManager().func_72375_a(aPlayer, tOriginalWorld);
			        aPlayer.playerNetServerHandler.setPlayerLocation(aX+0.5, aY+0.5, aZ+0.5, aPlayer.rotationYaw, aPlayer.rotationPitch);
			        aPlayer.theItemInWorldManager.setWorld(tTargetWorld);
			        MinecraftServer.getServer().getConfigurationManager().updateTimeAndWeatherForPlayer(aPlayer, tTargetWorld);
			        MinecraftServer.getServer().getConfigurationManager().syncPlayerInventory(aPlayer);
			        Iterator tIterator = aPlayer.getActivePotionEffects().iterator();
			        while (tIterator.hasNext()) {
			            PotionEffect potioneffect = (PotionEffect)tIterator.next();
			            aPlayer.playerNetServerHandler.sendPacket(new S1DPacketEntityEffect(aPlayer.getEntityId(), potioneffect));
			        }
			        aPlayer.playerNetServerHandler.setPlayerLocation(aX+0.5, aY+0.5, aZ+0.5, aPlayer.rotationYaw, aPlayer.rotationPitch);
			        FMLCommonHandler.instance().firePlayerChangedDimensionEvent(aPlayer, tOriginalWorld.provider.dimensionId, aDimension);
				} else {
					aEntity.setPosition(aX+0.5, aY+0.5, aZ+0.5);
					aEntity.worldObj.removeEntity(aEntity);
					aEntity.dimension = aDimension;
					aEntity.isDead = F;
		            Entity tNewEntity = EntityList.createEntityByName(EntityList.getEntityString(aEntity), tTargetWorld);
		            if (tNewEntity != null) {
		            	tNewEntity.copyDataFrom(aEntity, T);
			            aEntity.setDead();
		            	tNewEntity.isDead = F;
		            	boolean temp = tNewEntity.forceSpawn;
			            tNewEntity.forceSpawn = T;
		            	tTargetWorld.spawnEntityInWorld(tNewEntity);
			            tNewEntity.forceSpawn = temp;
		            	tNewEntity.isDead = F;
		            	aEntity = tNewEntity;
		            }
				}
				
				if (aEntity instanceof EntityLivingBase) {
					((EntityLivingBase)aEntity).setPositionAndUpdate(aX, aY, aZ);
				} else {
					aEntity.setPosition(aX, aY, aZ);
				}
				
	            tOriginalWorld.resetUpdateEntityTick();
	            tTargetWorld.resetUpdateEntityTick();
				return T;
			}
			return F;
		}
		
		/** to get a TileEntity properly, according to my additional Interfaces. Normally you should set aIgnoreUnloadedChunks to false, unless you have already checked these Coordinates, or you want to load Chunks */
	    public static DelegatorTileEntity<TileEntity> getTileEntity(World aWorld, int aX, int aY, int aZ, byte aSide, boolean aIgnoreUnloadedChunks) {
	    	TileEntity aTileEntity = getTileEntity(aWorld, aX, aY, aZ, aIgnoreUnloadedChunks);
			return aTileEntity instanceof ITileEntityDelegating ? ((ITileEntityDelegating)aTileEntity).getDelegateTileEntity(aSide) : new DelegatorTileEntity<TileEntity>(aTileEntity, aWorld, aX+OFFSETS_X[aSide], aY+OFFSETS_Y[aSide], aZ+OFFSETS_Z[aSide], aSide);
	    }
	    
		/** to get a TileEntity properly, according to my additional Interfaces. Normally you should set aIgnoreUnloadedChunks to false, unless you have already checked these Coordinates, or you want to load Chunks */
	    public static TileEntity getTileEntity(World aWorld, int aX, int aY, int aZ, boolean aIgnoreUnloadedChunks) {
	    	if (aIgnoreUnloadedChunks || aWorld.blockExists(aX, aY, aZ)) {
	        	TileEntity rTileEntity = aWorld.getTileEntity(aX, aY, aZ);
	        	if (rTileEntity != null) return rTileEntity;
	        	Block tBlock = aWorld.getBlock(aX, aY, aZ);
	        	return tBlock instanceof IBlockTileEntity ? ((IBlockTileEntity)tBlock).getTileEntity(aWorld, aX, aY, aZ) : null;
	    	}
			return null;
	    }
	    
		/** Sets the TileEntity at the passed position, with the option of turning adjacent TileEntity updates off. */
	    public static TileEntity setTileEntity(World aWorld, int aX, int aY, int aZ, TileEntity aTileEntity, boolean aCauseTileEntityUpdates) {
	    	if (aCauseTileEntityUpdates) aWorld.setTileEntity(aX, aY, aZ, aTileEntity); else {
		        Chunk tChunk = aWorld.getChunkFromChunkCoords(aX >> 4, aZ >> 4);
		        if (tChunk != null) {
		    		aWorld.addTileEntity(aTileEntity);
		        	tChunk.func_150812_a(aX & 15, aY, aZ & 15, aTileEntity);
		        	tChunk.isModified = true;
		        }
			}
	    	return aTileEntity;
	    }
	    
		/** @return the regular Temperature of the World at this Location according to my calculations. In Kelvin, ofcourse. */
		public static long getEnvironmentalTemperature(World aWorld, int aX, int aY, int aZ) {
			BiomeGenBase tBiome = aWorld.getBiomeGenForCoords(aX, aZ);
			return tBiome == null ? C + 20 : (long)(tBiome.temperature <= 0.2 ? tBiome.temperature * 5 * C : C + tBiome.temperature * 20);
		}
		
		/** @return the regular Temperature of the World at this Location according to my calculations. In Kelvin, ofcourse. */
		public static long getTemperature(World aWorld, int aX, int aY, int aZ) {
			long rTemperature = getEnvironmentalTemperature(aWorld, aX, aY, aZ);
			if (getCoordsOnFire(aWorld, aX, aY, aZ)) rTemperature = Math.max(rTemperature, C + 200);
			for (ChunkCoordinates tCoords : new ChunkCoordinates[] {new ChunkCoordinates(aX, aY, aZ), new ChunkCoordinates(aX+1, aY, aZ), new ChunkCoordinates(aX-1, aY, aZ), new ChunkCoordinates(aX, aY+1, aZ), new ChunkCoordinates(aX, aY-1, aZ), new ChunkCoordinates(aX, aY, aZ+1), new ChunkCoordinates(aX, aY, aZ-1)}) {
				Block tBlock = getBlock(aWorld, tCoords.posX, tCoords.posY, tCoords.posZ, F);
				if (tBlock == Blocks.lava || tBlock == Blocks.flowing_lava) rTemperature = Math.max(rTemperature, C + 500);
				else if (tBlock instanceof BlockFire) rTemperature = Math.max(rTemperature, C + 200);
			}
			return rTemperature;
		}
		
		public static ItemStack getStack(World aWorld, int aX, int aY, int aZ) {
			Block tBlock = aWorld.getBlock(aX, aY, aZ);
			return tBlock instanceof IBlockExtendedMetaData ? new ItemStack(tBlock, 1, ((IBlockExtendedMetaData)tBlock).getExtendedMetaData(aWorld, aX, aY, aZ)) : new ItemStack(tBlock, 1, aWorld.getBlockMetadata(aX, aY, aZ));
		}
		
		public static Block getBlock(World aWorld, int aX, int aY, int aZ, boolean aIgnoreUnloadedChunks) {
			if (aIgnoreUnloadedChunks || aWorld.blockExists(aX, aY, aZ)) return aWorld.getBlock(aX, aY, aZ);
			return Blocks.air;
		}
		
		public static boolean crossedChunkBorder(int aFromX, int aFromZ, int aToX, int aToZ) {
			return aFromX >> 4 != aToX >> 4 || aFromZ >> 4 != aToZ >> 4;
		}
		
		public static boolean setBlock(World aWorld, int aX, int aY, int aZ, ItemStack aStack) {
			Block tBlock = Stacks.block(aStack);
			if (tBlock instanceof IBlockPlacable) return ((IBlockPlacable)tBlock).placeBlock(aWorld, aX, aY, aZ, (byte)6, (short)aStack.getItemDamage(), aStack.getTagCompound(), true, false);
			if (tBlock != Blocks.air && aStack.getItemDamage() < 16) return aWorld.setBlock(aX, aY, aZ, tBlock, aStack.getItemDamage(), 3);
			return false;
		}
		
		public static boolean isBlockOccluded(World aWorld, int aX, int aY, int aZ, boolean aIgnoreUnloadedChunks, boolean aDefault) {
			return isBlockOpaque(aWorld, aX+1, aY, aZ, aIgnoreUnloadedChunks, aDefault) && isBlockOpaque(aWorld, aX-1, aY, aZ, aIgnoreUnloadedChunks, aDefault) && isBlockOpaque(aWorld, aX, aY+1, aZ, T, aDefault) && isBlockOpaque(aWorld, aX, aY-1, aZ, T, aDefault) && isBlockOpaque(aWorld, aX, aY, aZ+1, aIgnoreUnloadedChunks, aDefault) && isBlockOpaque(aWorld, aX, aY, aZ-1, aIgnoreUnloadedChunks, aDefault);
		}
		
		public static boolean isBlockOpaque(World aWorld, int aX, int aY, int aZ, boolean aIgnoreUnloadedChunks, boolean aDefault) {
			if (aIgnoreUnloadedChunks || aWorld.blockExists(aX, aY, aZ)) return aWorld.getBlock(aX, aY, aZ).isOpaqueCube();
			return aDefault;
		}
		
		public static boolean isBlockAir(World aWorld, int aX, int aY, int aZ) {
			return aWorld.getBlock(aX, aY, aZ).isAir(aWorld, aX, aY, aZ);
		}
		
		public static boolean hasBlockHitBox(World aWorld, int aX, int aY, int aZ) {
			return aWorld.getBlock(aX, aY, aZ).getCollisionBoundingBoxFromPool(aWorld, aX, aY, aZ) != null;
		}
		
		public static void setOnFire(World aWorld, int aX, int aY, int aZ, boolean aReplaceCenter, boolean aCheckFlammability) {
			if (aReplaceCenter)
			setToFire(aWorld, aX  , aY  , aZ  , aCheckFlammability);
			setToFire(aWorld, aX+1, aY  , aZ  , aCheckFlammability);
			setToFire(aWorld, aX-1, aY  , aZ  , aCheckFlammability);
			setToFire(aWorld, aX  , aY+1, aZ  , aCheckFlammability);
			setToFire(aWorld, aX  , aY-1, aZ  , aCheckFlammability);
			setToFire(aWorld, aX  , aY  , aZ+1, aCheckFlammability);
			setToFire(aWorld, aX  , aY  , aZ-1, aCheckFlammability);
		}
		
		public static boolean setToFire(World aWorld, int aX, int aY, int aZ, boolean aCheckFlammability) {
			Block tBlock = aWorld.getBlock(aX, aY, aZ);
			if (tBlock.getMaterial() == Material.carpet || tBlock.getCollisionBoundingBoxFromPool(aWorld, aX, aY, aZ) == null) {
				if (tBlock.getFlammability(aWorld, aX, aY, aZ, FORGE_DIR[SIDE_ANY]) > 0) return aWorld.setBlock(aX, aY, aZ, Blocks.fire);
				if (aCheckFlammability) {
					for (byte tSide : ALL_SIDES_VALID) if (aWorld.getBlock(aX+OFFSETS_X[tSide], aY+OFFSETS_Y[tSide], aZ+OFFSETS_Z[tSide]).getFlammability(aWorld, aX+OFFSETS_X[tSide], aY+OFFSETS_Y[tSide], aZ+OFFSETS_Z[tSide], FORGE_DIR_OPPOSITES[tSide]) > 0) return aWorld.setBlock(aX, aY, aZ, Blocks.fire);
				} else {
					return aWorld.setBlock(aX, aY, aZ, Blocks.fire);
				}
			}
			return false;
		}
		
		public static boolean getCoordsOnFire(World aWorld, int aX, int aY, int aZ) {
			return getBlock(aWorld, aX, aY, aZ, F) instanceof BlockFire || getBlock(aWorld, aX+1, aY, aZ, F) instanceof BlockFire || getBlock(aWorld, aX-1, aY, aZ, F) instanceof BlockFire || getBlock(aWorld, aX, aY+1, aZ, F) instanceof BlockFire || getBlock(aWorld, aX, aY-1, aZ, F) instanceof BlockFire || getBlock(aWorld, aX, aY, aZ+1, F) instanceof BlockFire || getBlock(aWorld, aX, aY, aZ-1, F) instanceof BlockFire;
		}
		
		public static int getCoordinateScan(ArrayList<String> aList, EntityPlayer aPlayer, World aWorld, int aScanLevel, int aX, int aY, int aZ, int aSide, float aClickX, float aClickY, float aClickZ) {
			if (aList == null) return 0;
			
			ArrayList<String> tList = new ArrayListNoNulls<String>();
			int rEUAmount = 0;
			
			TileEntity tTileEntity = getTileEntity(aWorld, aX, aY, aZ, true);
		    
		    Block tBlock = aWorld.getBlock(aX, aY, aZ);
		    
		    tList.add("----- X: " + aX + " Y: " + aY + " Z: " + aZ + " -----");
		    try {
			    if (tTileEntity != null && tTileEntity instanceof IInventory)
			    	tList.add("Name: " + ((IInventory)tTileEntity).getInventoryName() + "  MetaData: " + aWorld.getBlockMetadata(aX, aY, aZ));
			    else
			    	tList.add("Name: " + tBlock.getUnlocalizedName() + "  MetaData: " + aWorld.getBlockMetadata(aX, aY, aZ));
			    
			    tList.add("Hardness: " + tBlock.getBlockHardness(aWorld, aX, aY, aZ) + "  Blast Resistance: " + tBlock.getExplosionResistance(aPlayer, aWorld, aX, aY, aZ, aPlayer.posX, aPlayer.posY, aPlayer.posZ));
			    if (tBlock.isBeaconBase(aWorld, aX, aY, aZ, aX, aY+1, aZ)) tList.add("Is valid Beacon Pyramid Material");
			} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
		    if (tTileEntity != null) {
				try {if (tTileEntity instanceof IFluidHandler) {
					rEUAmount+=500;
				    FluidTankInfo[] tTanks = ((IFluidHandler)tTileEntity).getTankInfo(FORGE_DIR[aSide]);
				    if (tTanks != null) for (byte i = 0; i < tTanks.length; i++) {
				    	tList.add("Tank " + i + ": " + (tTanks[i].fluid==null?0:tTanks[i].fluid.amount) + " / " + tTanks[i].capacity + " " + Fluids.name(tTanks[i].fluid, T));
				    }
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.reactor.IReactorChamber) {
					rEUAmount+=500;
				    tTileEntity = (TileEntity)(((ic2.api.reactor.IReactorChamber)tTileEntity).getReactor());
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.reactor.IReactor) {
					rEUAmount+=500;
					tList.add("Heat: " + ((ic2.api.reactor.IReactor)tTileEntity).getHeat() + "/" + ((ic2.api.reactor.IReactor)tTileEntity).getMaxHeat()
							+ "  HEM: " + ((ic2.api.reactor.IReactor)tTileEntity).getHeatEffectModifier() + "  Base EU Output: "/* + ((ic2.api.reactor.IReactor)tTileEntity).getOutput()*/);
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.tile.IWrenchable) {
					rEUAmount+=100;
			        tList.add("Facing: " + ((ic2.api.tile.IWrenchable)tTileEntity).getFacing() + " / Chance: " + (((ic2.api.tile.IWrenchable)tTileEntity).getWrenchDropRate()*100) + "%");
				    tList.add(((ic2.api.tile.IWrenchable)tTileEntity).wrenchCanRemove(aPlayer)?"You can remove this with a Wrench":"You can NOT remove this with a Wrench");
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.energy.tile.IEnergyTile) {
					rEUAmount+=200;
				    //aList.add(((ic2.api.energy.tile.IEnergyTile)tTileEntity).isAddedToEnergyNet()?"Added to E-net":"Not added to E-net! Bug?");
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.energy.tile.IEnergySink) {
					rEUAmount+=400;
			        //aList.add("Demanded Energy: " + ((ic2.api.energy.tile.IEnergySink)tTileEntity).demandsEnergy());
			        //tList.add("Max Safe Input: " + getTier(((ic2.api.energy.tile.IEnergySink)tTileEntity).getSinkTier()));
			        //tList.add("Max Safe Input: " + ((ic2.api.energy.tile.IEnergySink)tTileEntity).getMaxSafeInput());
			    }} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.energy.tile.IEnergySource) {
					rEUAmount+=400;
			        //aList.add("Max Energy Output: " + ((ic2.api.energy.tile.IEnergySource)tTileEntity).getMaxEnergyOutput());
			    }} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.energy.tile.IEnergyConductor) {
					rEUAmount+=200;
			        tList.add("Conduction Loss: " + ((ic2.api.energy.tile.IEnergyConductor)tTileEntity).getConductionLoss());
			    }} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.tile.IEnergyStorage) {
					rEUAmount+=200;
			        tList.add("Contained Energy: " + ((ic2.api.tile.IEnergyStorage)tTileEntity).getStored() + " of " + ((ic2.api.tile.IEnergyStorage)tTileEntity).getCapacity());
			        //aList.add(((ic2.api.tile.IEnergyStorage)tTileEntity).isTeleporterCompatible(ic2.api.Direction.YP)?"Teleporter Compatible":"Not Teleporter Compatible");
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof IUpgradableMachine) {
					rEUAmount+=500;
			    	if (((IUpgradableMachine)tTileEntity).hasMufflerUpgrade()) tList.add("Has Muffler Upgrade");
			    }} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof IMachineProgress) {
					rEUAmount+=400;
			    	int tValue = 0;
			    	if (0 < (tValue = ((IMachineProgress)tTileEntity).getMaxProgress())) tList.add("Progress: " + tValue + " / " + ((IMachineProgress)tTileEntity).getProgress());
			    }} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ICoverable) {
					rEUAmount+=300;
			    	String tString = ((ICoverable)tTileEntity).getCoverBehaviorAtSide((byte)aSide).getDescription((byte)aSide, ((ICoverable)tTileEntity).getCoverIDAtSide((byte)aSide), ((ICoverable)tTileEntity).getCoverDataAtSide((byte)aSide), (ICoverable)tTileEntity);
			    	if (tString != null && !tString.equals("")) tList.add(tString);
			    }} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof IBasicEnergyContainer && ((IBasicEnergyContainer)tTileEntity).getEUCapacity() > 0) {
					tList.add("Max IN: " + ((IBasicEnergyContainer)tTileEntity).getInputVoltage() + " EU");
				    tList.add("Max OUT: " + ((IBasicEnergyContainer)tTileEntity).getOutputVoltage() + " EU at " + ((IBasicEnergyContainer)tTileEntity).getOutputAmperage() + " Amperes");
					tList.add("Energy: " + ((IBasicEnergyContainer)tTileEntity).getStoredEU() + " / " + ((IBasicEnergyContainer)tTileEntity).getEUCapacity() + "EU");
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof IGregTechTileEntity) {
			    	tList.add("Owned by: " + ((IGregTechTileEntity)tTileEntity).getOwnerName());
			    }} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof IGregTechDeviceInformation && ((IGregTechDeviceInformation)tTileEntity).isGivingInformation()) {
					tList.addAll(Arrays.asList(((IGregTechDeviceInformation)tTileEntity).getInfoData()));
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
				try {if (tTileEntity instanceof ic2.api.crops.ICropTile) {
					if (((ic2.api.crops.ICropTile)tTileEntity).getScanLevel() < 4) {
						rEUAmount+=10000;
						((ic2.api.crops.ICropTile)tTileEntity).setScanLevel((byte)4);
					}
					rEUAmount+=1000;
					tList.add("Type -- Crop-Name: " + LanguageHandler.translate(((ic2.api.crops.ICropTile)tTileEntity).getCrop().name())
				    		+ "  Growth: " + ((ic2.api.crops.ICropTile)tTileEntity).getGrowth()
				    		+ "  Gain: " + ((ic2.api.crops.ICropTile)tTileEntity).getGain()
				    		+ "  Resistance: " + ((ic2.api.crops.ICropTile)tTileEntity).getResistance()
				    		);
				    tList.add("Plant -- Fertilizer: " + ((ic2.api.crops.ICropTile)tTileEntity).getNutrientStorage()
				    		+ "  Water: " + ((ic2.api.crops.ICropTile)tTileEntity).getHydrationStorage()
				    		+ "  Weed-Ex: " + ((ic2.api.crops.ICropTile)tTileEntity).getWeedExStorage()
				    		+ "  Scan-Level: " + ((ic2.api.crops.ICropTile)tTileEntity).getScanLevel()
				    		);
				    tList.add("Environment -- Nutrients: " + ((ic2.api.crops.ICropTile)tTileEntity).getNutrients()
				    		+ "  Humidity: " + ((ic2.api.crops.ICropTile)tTileEntity).getHumidity()
				    		+ "  Air-Quality: " + ((ic2.api.crops.ICropTile)tTileEntity).getAirQuality()
				    		);
				    String tString = "";
				    for (String tAttribute : ((ic2.api.crops.ICropTile)tTileEntity).getCrop().attributes()) {
				    	tString += ", " + tAttribute;
				    }
				    tList.add("Attributes:" + tString.replaceFirst(",", ""));
				    tList.add("Discovered by: " + ((ic2.api.crops.ICropTile)tTileEntity).getCrop().discoveredBy());
					
				}} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
	    	}
		    try {if (tBlock instanceof IDebugableBlock) {
		    	rEUAmount+=500;
		        ArrayList<String> temp = ((IDebugableBlock)tBlock).getDebugInfo(aPlayer, aX, aY, aZ, 3);
		        if (temp != null) tList.addAll(temp);
		    }} catch(Throwable e) {if (D1) e.printStackTrace(GT_Log.err);}
		    
		    BlockScanningEvent tEvent = new BlockScanningEvent(aWorld, aPlayer, aX, aY, aZ, (byte)aSide, aScanLevel, tBlock, tTileEntity, tList, aClickX, aClickY, aClickZ);
		    tEvent.mEUCost = rEUAmount;
		    MinecraftForge.EVENT_BUS.post(tEvent);
		    if (!tEvent.isCanceled()) aList.addAll(tList);
			return tEvent.mEUCost;
		}
	}
	
	public static synchronized boolean removeSimpleIC2MachineRecipe(ItemStack aInput, Map<IRecipeInput, RecipeOutput> aRecipeList, ItemStack aOutput) {
		if ((Stacks.invalid(aInput) && Stacks.invalid(aOutput)) || aRecipeList == null) return F;
		boolean rReturn = F;
		Iterator<Map.Entry<IRecipeInput, RecipeOutput>> tIterator = aRecipeList.entrySet().iterator();
		aOutput = OM.get(aOutput);
		while (tIterator.hasNext()) {
			Map.Entry<IRecipeInput, RecipeOutput> tEntry = tIterator.next();
			if (aInput == null || tEntry.getKey().matches(aInput)) {
				List<ItemStack> tList = tEntry.getValue().items;
				if (tList != null) for (ItemStack tOutput : tList) if (aOutput == null || Stacks.equal(OM.get(tOutput), aOutput)) {
					tIterator.remove();
					rReturn = T;
					break;
				}
			}
		}
		return rReturn;
	}
	
	public static boolean addSimpleIC2MachineRecipe(ItemStack aInput, Map<IRecipeInput, RecipeOutput> aRecipeList, NBTTagCompound aNBT, Object... aOutput) {
		if (Stacks.invalid(aInput) || aOutput.length == 0 || aRecipeList == null) return F;
		OreDictItemData tOreName = OM.association(aInput);
		if (tOreName != null) {
			aRecipeList.put(new RecipeInputOreDict(tOreName.toString(), aInput.stackSize), new RecipeOutput(aNBT, OreDictManager.INSTANCE.getStackArray(T, aOutput)));
		} else {
			aRecipeList.put(new RecipeInputItemStack(Stacks.copy(aInput), aInput.stackSize), new RecipeOutput(aNBT, OreDictManager.INSTANCE.getStackArray(T, aOutput)));
		}
		return T;
	}
}